from apscheduler.schedulers.base import BaseScheduler
from apscheduler.schedulers.blocking import BlockingScheduler

try:
    import gevent
    from gevent.event import Event
    from gevent.lock import RLock
except ImportError as exc:  # pragma: nocover
    raise ImportError("GeventScheduler requires gevent installed") from exc


class GeventScheduler(BlockingScheduler):
    """A scheduler that runs as a Gevent greenlet."""

    _greenlet = None

    def start(self, *args, **kwargs):
        self._event = Event()
        BaseScheduler.start(self, *args, **kwargs)
        self._greenlet = gevent.spawn(self._main_loop)
        return self._greenlet

    def shutdown(self, *args, **kwargs):
        super().shutdown(*args, **kwargs)
        self._greenlet.join()
        del self._greenlet

    def _create_lock(self):
        return RLock()

    def _create_default_executor(self):
        from apscheduler.executors.gevent import GeventExecutor

        return GeventExecutor()



"""
Demonstrates how to use the asyncio compatible scheduler to schedule a job that executes on 3
second intervals.
"""

import asyncio
import os
from datetime import datetime

from apscheduler.schedulers.asyncio import AsyncIOScheduler


def tick():
    print(f"Tick! The time is: {datetime.now()}")


async def main():
    scheduler = AsyncIOScheduler()
    scheduler.add_job(tick, "interval", seconds=3)
    scheduler.start()
    print("Press Ctrl+{} to exit".format("Break" if os.name == "nt" else "C"))
    while True:
        await asyncio.sleep(1000)


if __name__ == "__main__":
    # Execution will block here until Ctrl+C (Ctrl+Break on Windows) is pressed.
    try:
        asyncio.run(main())
    except (KeyboardInterrupt, SystemExit):
        pass



#!/usr/bin/env python3
from __future__ import annotations

from importlib.metadata import version as get_version

from packaging.version import parse

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.intersphinx",
    "sphinx_rtd_theme",
]

templates_path = ["_templates"]
source_suffix = ".rst"
master_doc = "index"
project = "APScheduler"
author = "Alex Grönholm"
copyright = "2009, " + author

v = parse(get_version("APScheduler"))
version = v.base_version
release = v.public

language = "en"

exclude_patterns = ["_build"]
pygments_style = "sphinx"
autodoc_default_options = {"members": True, "show-inheritance": True}
autodoc_mock_imports = [
    "bson",
    "gevent",
    "kazoo",
    "pymongo",
    "PyQt6",
    "redis",
    "rethinkdb",
    "sqlalchemy",
    "tornado",
    "twisted",
]
todo_include_todos = False
html_theme = "sphinx_rtd_theme"

intersphinx_mapping = {
    "python": ("https://docs.python.org/3/", None),
    "sqlalchemy": ("https://docs.sqlalchemy.org/en/20/", None),
}



from datetime import datetime, timedelta

from tzlocal import get_localzone

from apscheduler.triggers.base import BaseTrigger
from apscheduler.triggers.cron.fields import (
    DEFAULT_VALUES,
    BaseField,
    DayOfMonthField,
    DayOfWeekField,
    MonthField,
    WeekField,
)
from apscheduler.util import (
    astimezone,
    convert_to_datetime,
    datetime_ceil,
    datetime_repr,
)


class CronTrigger(BaseTrigger):
    """
    Triggers when current time matches all specified time constraints,
    similarly to how the UNIX cron scheduler works.

    :param int|str year: 4-digit year
    :param int|str month: month (1-12)
    :param int|str day: day of month (1-31)
    :param int|str week: ISO week (1-53)
    :param int|str day_of_week: number or name of weekday (0-6 or mon,tue,wed,thu,fri,sat,sun)
    :param int|str hour: hour (0-23)
    :param int|str minute: minute (0-59)
    :param int|str second: second (0-59)
    :param datetime|str start_date: earliest possible date/time to trigger on (inclusive)
    :param datetime|str end_date: latest possible date/time to trigger on (inclusive)
    :param datetime.tzinfo|str timezone: time zone to use for the date/time calculations (defaults
        to scheduler timezone)
    :param int|None jitter: delay the job execution by ``jitter`` seconds at most

    .. note:: The first weekday is always **monday**.
    """

    FIELD_NAMES = (
        "year",
        "month",
        "day",
        "week",
        "day_of_week",
        "hour",
        "minute",
        "second",
    )
    FIELDS_MAP = {
        "year": BaseField,
        "month": MonthField,
        "week": WeekField,
        "day": DayOfMonthField,
        "day_of_week": DayOfWeekField,
        "hour": BaseField,
        "minute": BaseField,
        "second": BaseField,
    }

    __slots__ = "timezone", "start_date", "end_date", "fields", "jitter"

    def __init__(
        self,
        year=None,
        month=None,
        day=None,
        week=None,
        day_of_week=None,
        hour=None,
        minute=None,
        second=None,
        start_date=None,
        end_date=None,
        timezone=None,
        jitter=None,
    ):
        if timezone:
            self.timezone = astimezone(timezone)
        elif isinstance(start_date, datetime) and start_date.tzinfo:
            self.timezone = astimezone(start_date.tzinfo)
        elif isinstance(end_date, datetime) and end_date.tzinfo:
            self.timezone = astimezone(end_date.tzinfo)
        else:
            self.timezone = get_localzone()

        self.start_date = convert_to_datetime(start_date, self.timezone, "start_date")
        self.end_date = convert_to_datetime(end_date, self.timezone, "end_date")

        self.jitter = jitter

        values = dict(
            (key, value)
            for (key, value) in locals().items()
            if key in self.FIELD_NAMES and value is not None
        )
        self.fields = []
        assign_defaults = False
        for field_name in self.FIELD_NAMES:
            if field_name in values:
                exprs = values.pop(field_name)
                is_default = False
                assign_defaults = not values
            elif assign_defaults:
                exprs = DEFAULT_VALUES[field_name]
                is_default = True
            else:
                exprs = "*"
                is_default = True

            field_class = self.FIELDS_MAP[field_name]
            field = field_class(field_name, exprs, is_default)
            self.fields.append(field)

    @classmethod
    def from_crontab(cls, expr, timezone=None):
        """
        Create a :class:`~CronTrigger` from a standard crontab expression.

        See https://en.wikipedia.org/wiki/Cron for more information on the format accepted here.

        :param expr: minute, hour, day of month, month, day of week
        :param datetime.tzinfo|str timezone: time zone to use for the date/time calculations (
            defaults to scheduler timezone)
        :return: a :class:`~CronTrigger` instance

        """
        values = expr.split()
        if len(values) != 5:
            raise ValueError(f"Wrong number of fields; got {len(values)}, expected 5")

        return cls(
            minute=values[0],
            hour=values[1],
            day=values[2],
            month=values[3],
            day_of_week=values[4],
            timezone=timezone,
        )

    def _increment_field_value(self, dateval, fieldnum):
        """
        Increments the designated field and resets all less significant fields to their minimum
        values.

        :type dateval: datetime
        :type fieldnum: int
        :return: a tuple containing the new date, and the number of the field that was actually
            incremented
        :rtype: tuple
        """

        values = {}
        i = 0
        while i < len(self.fields):
            field = self.fields[i]
            if not field.REAL:
                if i == fieldnum:
                    fieldnum -= 1
                    i -= 1
                else:
                    i += 1
                continue

            if i < fieldnum:
                values[field.name] = field.get_value(dateval)
                i += 1
            elif i > fieldnum:
                values[field.name] = field.get_min(dateval)
                i += 1
            else:
                value = field.get_value(dateval)
                maxval = field.get_max(dateval)
                if value == maxval:
                    fieldnum -= 1
                    i -= 1
                else:
                    values[field.name] = value + 1
                    i += 1

        difference = datetime(**values) - dateval.replace(tzinfo=None)
        dateval = datetime.fromtimestamp(
            dateval.timestamp() + difference.total_seconds(), self.timezone
        )
        return dateval, fieldnum

    def _set_field_value(self, dateval, fieldnum, new_value):
        values = {}
        for i, field in enumerate(self.fields):
            if field.REAL:
                if i < fieldnum:
                    values[field.name] = field.get_value(dateval)
                elif i > fieldnum:
                    values[field.name] = field.get_min(dateval)
                else:
                    values[field.name] = new_value

        return datetime(**values, tzinfo=self.timezone, fold=dateval.fold)

    def get_next_fire_time(self, previous_fire_time, now):
        if previous_fire_time:
            start_date = min(now, previous_fire_time + timedelta(microseconds=1))
            if start_date == previous_fire_time:
                start_date += timedelta(microseconds=1)
        else:
            start_date = max(now, self.start_date) if self.start_date else now

        fieldnum = 0
        next_date = datetime_ceil(start_date).astimezone(self.timezone)
        while 0 <= fieldnum < len(self.fields):
            field = self.fields[fieldnum]
            curr_value = field.get_value(next_date)
            next_value = field.get_next_value(next_date)

            if next_value is None:
                # No valid value was found
                next_date, fieldnum = self._increment_field_value(
                    next_date, fieldnum - 1
                )
            elif next_value > curr_value:
                # A valid, but higher than the starting value, was found
                if field.REAL:
                    next_date = self._set_field_value(next_date, fieldnum, next_value)
                    fieldnum += 1
                else:
                    next_date, fieldnum = self._increment_field_value(
                        next_date, fieldnum
                    )
            else:
                # A valid value was found, no changes necessary
                fieldnum += 1

            # Return if the date has rolled past the end date
            if self.end_date and next_date > self.end_date:
                return None

        if fieldnum >= 0:
            next_date = self._apply_jitter(next_date, self.jitter, now)
            return min(next_date, self.end_date) if self.end_date else next_date

    def __getstate__(self):
        return {
            "version": 2,
            "timezone": self.timezone,
            "start_date": self.start_date,
            "end_date": self.end_date,
            "fields": self.fields,
            "jitter": self.jitter,
        }

    def __setstate__(self, state):
        # This is for compatibility with APScheduler 3.0.x
        if isinstance(state, tuple):
            state = state[1]

        if state.get("version", 1) > 2:
            raise ValueError(
                f"Got serialized data for version {state['version']} of "
                f"{self.__class__.__name__}, but only versions up to 2 can be handled"
            )

        self.timezone = astimezone(state["timezone"])
        self.start_date = state["start_date"]
        self.end_date = state["end_date"]
        self.fields = state["fields"]
        self.jitter = state.get("jitter")

    def __str__(self):
        options = [f"{f.name}='{f}'" for f in self.fields if not f.is_default]
        return "cron[{}]".format(", ".join(options))

    def __repr__(self):
        options = [f"{f.name}='{f}'" for f in self.fields if not f.is_default]
        if self.start_date:
            options.append(f"start_date={datetime_repr(self.start_date)!r}")
        if self.end_date:
            options.append(f"end_date={datetime_repr(self.end_date)!r}")
        if self.jitter:
            options.append(f"jitter={self.jitter}")

        return "<{} ({}, timezone='{}')>".format(
            self.__class__.__name__,
            ", ".join(options),
            self.timezone,
        )



"""
This is an example RPC client that connects to the RPyC based scheduler service.

It first connects to the RPyC server on localhost:12345.
Then it schedules a job to run on 2 second intervals and sleeps for 10 seconds.
After that, it unschedules the job and exits.
"""

from time import sleep

import rpyc

conn = rpyc.connect("localhost", 12345)
job = conn.root.add_job(
    "server:print_text", "interval", args=["Hello, World"], seconds=2
)
sleep(10)
conn.root.remove_job(job.id)



import pickle
import sys
from datetime import datetime, timedelta
from unittest.mock import Mock

import pytest
import pytz

from apscheduler.triggers.cron import CronTrigger
from apscheduler.util import localize

if sys.version_info < (3, 9):
    from backports.zoneinfo import ZoneInfo
else:
    from zoneinfo import ZoneInfo


def test_cron_trigger_1(timezone):
    trigger = CronTrigger(year="2009/2", month="1/3", day="5-13", timezone=timezone)
    assert repr(trigger) == (
        "<CronTrigger (year='2009/2', month='1/3', day='5-13', "
        "timezone='Europe/Berlin')>"
    )
    assert str(trigger) == "cron[year='2009/2', month='1/3', day='5-13']"
    start_date = localize(datetime(2008, 12, 1), timezone)
    correct_next_date = localize(datetime(2009, 1, 5), timezone)
    assert trigger.get_next_fire_time(None, start_date) == correct_next_date


def test_cron_trigger_2(timezone):
    trigger = CronTrigger(year="2009/2", month="1/3", day="5-13", timezone=timezone)
    start_date = localize(datetime(2009, 10, 14), timezone)
    correct_next_date = localize(datetime(2011, 1, 5), timezone)
    assert trigger.get_next_fire_time(None, start_date) == correct_next_date


def test_cron_trigger_3(timezone):
    trigger = CronTrigger(year="2009", month="feb-dec", hour="8-10", timezone=timezone)
    assert repr(trigger) == (
        "<CronTrigger (year='2009', month='feb-dec', hour='8-10', "
        "timezone='Europe/Berlin')>"
    )
    start_date = localize(datetime(2009, 1, 1), timezone)
    correct_next_date = localize(datetime(2009, 2, 1, 8), timezone)
    assert trigger.get_next_fire_time(None, start_date) == correct_next_date


def test_cron_trigger_4(timezone):
    trigger = CronTrigger(year="2012", month="2", day="last", timezone=timezone)
    assert repr(trigger) == (
        "<CronTrigger (year='2012', month='2', day='last', "
        "timezone='Europe/Berlin')>"
    )
    start_date = localize(datetime(2012, 2, 1), timezone)
    correct_next_date = localize(datetime(2012, 2, 29), timezone)
    assert trigger.get_next_fire_time(None, start_date) == correct_next_date


def test_start_end_times_string(timezone, monkeypatch):
    monkeypatch.setattr(
        "apscheduler.triggers.cron.get_localzone", Mock(return_value=timezone)
    )
    trigger = CronTrigger(
        start_date="2016-11-05 05:06:53", end_date="2017-11-05 05:11:32"
    )
    assert trigger.start_date == localize(datetime(2016, 11, 5, 5, 6, 53), timezone)
    assert trigger.end_date == localize(datetime(2017, 11, 5, 5, 11, 32), timezone)


def test_cron_zero_value(timezone):
    trigger = CronTrigger(year=2009, month=2, hour=0, timezone=timezone)
    assert repr(trigger) == (
        "<CronTrigger (year='2009', month='2', hour='0', timezone='Europe/Berlin')>"
    )


def test_cron_year_list(timezone):
    trigger = CronTrigger(year="2009,2008", timezone=timezone)
    assert repr(trigger) == "<CronTrigger (year='2009,2008', timezone='Europe/Berlin')>"
    assert str(trigger) == "cron[year='2009,2008']"
    start_date = localize(datetime(2009, 1, 1), timezone)
    correct_next_date = localize(datetime(2009, 1, 1), timezone)
    assert trigger.get_next_fire_time(None, start_date) == correct_next_date


def test_cron_start_date(timezone):
    trigger = CronTrigger(
        year="2009",
        month="2",
        hour="8-10",
        start_date="2009-02-03 11:00:00",
        timezone=timezone,
    )
    assert repr(trigger) == (
        "<CronTrigger (year='2009', month='2', hour='8-10', "
        "start_date='2009-02-03 11:00:00 CET', "
        "timezone='Europe/Berlin')>"
    )
    assert str(trigger) == "cron[year='2009', month='2', hour='8-10']"
    start_date = localize(datetime(2009, 1, 1), timezone)
    correct_next_date = localize(datetime(2009, 2, 4, 8), timezone)
    assert trigger.get_next_fire_time(None, start_date) == correct_next_date


def test_previous_fire_time_1(timezone):
    """Test for previous_fire_time arg in get_next_fire_time()"""
    trigger = CronTrigger(day="*", timezone=timezone)
    previous_fire_time = localize(datetime(2015, 11, 23), timezone)
    now = localize(datetime(2015, 11, 26), timezone)
    correct_next_date = localize(datetime(2015, 11, 24), timezone)
    assert trigger.get_next_fire_time(previous_fire_time, now) == correct_next_date


def test_previous_fire_time_2(timezone):
    trigger = CronTrigger(day="*", timezone=timezone)
    previous_fire_time = localize(datetime(2015, 11, 23), timezone)
    now = localize(datetime(2015, 11, 22), timezone)
    correct_next_date = localize(datetime(2015, 11, 22), timezone)
    assert trigger.get_next_fire_time(previous_fire_time, now) == correct_next_date


def test_previous_fire_time_3(timezone):
    trigger = CronTrigger(day="*", timezone=timezone)
    previous_fire_time = localize(datetime(2016, 4, 25), timezone)
    now = localize(datetime(2016, 4, 25), timezone)
    correct_next_date = localize(datetime(2016, 4, 26), timezone)
    assert trigger.get_next_fire_time(previous_fire_time, now) == correct_next_date


def test_cron_weekday_overlap(timezone):
    trigger = CronTrigger(
        year=2009, month=1, day="6-10", day_of_week="2-4", timezone=timezone
    )
    assert repr(trigger) == (
        "<CronTrigger (year='2009', month='1', day='6-10', "
        "day_of_week='2-4', timezone='Europe/Berlin')>"
    )
    assert str(trigger) == "cron[year='2009', month='1', day='6-10', day_of_week='2-4']"
    start_date = localize(datetime(2009, 1, 1), timezone)
    correct_next_date = localize(datetime(2009, 1, 7), timezone)
    assert trigger.get_next_fire_time(None, start_date) == correct_next_date


def test_cron_weekday_nomatch(timezone):
    trigger = CronTrigger(
        year=2009, month=1, day="6-10", day_of_week="0,6", timezone=timezone
    )
    assert repr(trigger) == (
        "<CronTrigger (year='2009', month='1', day='6-10', "
        "day_of_week='0,6', timezone='Europe/Berlin')>"
    )
    assert str(trigger) == "cron[year='2009', month='1', day='6-10', day_of_week='0,6']"
    start_date = localize(datetime(2009, 1, 1), timezone)
    correct_next_date = None
    assert trigger.get_next_fire_time(None, start_date) == correct_next_date


def test_cron_weekday_positional(timezone):
    trigger = CronTrigger(year=2009, month=1, day="4th wed", timezone=timezone)
    assert repr(trigger) == (
        "<CronTrigger (year='2009', month='1', day='4th wed', "
        "timezone='Europe/Berlin')>"
    )
    assert str(trigger) == "cron[year='2009', month='1', day='4th wed']"
    start_date = localize(datetime(2009, 1, 1), timezone)
    correct_next_date = localize(datetime(2009, 1, 28), timezone)
    assert trigger.get_next_fire_time(None, start_date) == correct_next_date


def test_week_1(timezone):
    trigger = CronTrigger(year=2009, month=2, week=8, timezone=timezone)
    assert repr(trigger) == (
        "<CronTrigger (year='2009', month='2', week='8', timezone='Europe/Berlin')>"
    )
    assert str(trigger) == "cron[year='2009', month='2', week='8']"
    start_date = localize(datetime(2009, 1, 1), timezone)
    correct_next_date = localize(datetime(2009, 2, 16), timezone)
    assert trigger.get_next_fire_time(None, start_date) == correct_next_date


def test_week_2(timezone):
    trigger = CronTrigger(year=2009, week=15, day_of_week=2, timezone=timezone)
    assert repr(trigger) == (
        "<CronTrigger (year='2009', week='15', day_of_week='2', "
        "timezone='Europe/Berlin')>"
    )
    assert str(trigger) == "cron[year='2009', week='15', day_of_week='2']"
    start_date = localize(datetime(2009, 1, 1), timezone)
    correct_next_date = localize(datetime(2009, 4, 8), timezone)
    assert trigger.get_next_fire_time(None, start_date) == correct_next_date


def test_cron_extra_coverage(timezone):
    # This test has no value other than patching holes in test coverage
    trigger = CronTrigger(day="6,8", timezone=timezone)
    assert repr(trigger) == "<CronTrigger (day='6,8', timezone='Europe/Berlin')>"
    assert str(trigger) == "cron[day='6,8']"
    start_date = localize(datetime(2009, 12, 31), timezone)
    correct_next_date = localize(datetime(2010, 1, 6), timezone)
    assert trigger.get_next_fire_time(None, start_date) == correct_next_date


def test_cron_faulty_expr(timezone):
    pytest.raises(ValueError, CronTrigger, year="2009-fault", timezone=timezone)


def test_cron_increment_weekday(timezone):
    """
    Tests that incrementing the weekday field in the process of calculating the next matching
    date won't cause problems.

    """
    trigger = CronTrigger(hour="5-6", timezone=timezone)
    assert repr(trigger) == "<CronTrigger (hour='5-6', timezone='Europe/Berlin')>"
    assert str(trigger) == "cron[hour='5-6']"
    start_date = localize(datetime(2009, 9, 25, 7), timezone)
    correct_next_date = localize(datetime(2009, 9, 26, 5), timezone)
    assert trigger.get_next_fire_time(None, start_date) == correct_next_date


def test_cron_bad_kwarg(timezone):
    pytest.raises(TypeError, CronTrigger, second=0, third=1, timezone=timezone)


def test_month_rollover(timezone):
    trigger = CronTrigger(timezone=timezone, day=30)
    now = localize(datetime(2016, 2, 1), timezone)
    expected = localize(datetime(2016, 3, 30), timezone)
    assert trigger.get_next_fire_time(None, now) == expected


def test_timezone_from_start_date(timezone):
    """
    Tests that the trigger takes the timezone from the start_date parameter if no
    timezone is supplied.

    """
    start_date = localize(datetime(2014, 4, 13, 5, 30), timezone)
    trigger = CronTrigger(year=2014, hour=4, start_date=start_date)
    assert trigger.timezone.key == "Europe/Berlin"


def test_end_date(timezone):
    end_date = localize(datetime(2014, 4, 13, 3), timezone)
    trigger = CronTrigger(year=2014, hour=4, end_date=end_date)

    start_date = localize(datetime(2014, 4, 13, 2, 30), timezone)
    assert trigger.get_next_fire_time(
        None, start_date - timedelta(1)
    ) == start_date.replace(day=12, hour=4, minute=0)
    assert trigger.get_next_fire_time(None, start_date) is None


def test_different_tz(timezone):
    alter_tz = pytz.FixedOffset(-600)
    trigger = CronTrigger(year=2009, week=15, day_of_week=2, timezone=timezone)
    assert repr(trigger) == (
        "<CronTrigger (year='2009', week='15', day_of_week='2', "
        "timezone='Europe/Berlin')>"
    )
    assert str(trigger) == "cron[year='2009', week='15', day_of_week='2']"
    start_date = alter_tz.localize(datetime(2008, 12, 31, 22))
    correct_next_date = localize(datetime(2009, 4, 8), timezone)
    assert trigger.get_next_fire_time(None, start_date) == correct_next_date


@pytest.mark.parametrize(
    "trigger_args, start_date, start_date_fold, correct_next_date",
    [
        ({"hour": 8}, datetime(2013, 3, 9, 12), False, datetime(2013, 3, 10, 8)),
        ({"hour": 8}, datetime(2013, 11, 2, 12), True, datetime(2013, 11, 3, 8)),
        (
            {"minute": "*/30"},
            datetime(2013, 3, 10, 1, 35),
            1,
            datetime(2013, 3, 10, 3),
        ),
        (
            {"minute": "*/30"},
            datetime(2013, 11, 3, 1, 35),
            0,
            datetime(2013, 11, 3, 1),
        ),
    ],
    ids=[
        "absolute_spring",
        "absolute_autumn",
        "interval_spring",
        "interval_autumn",
    ],
)
def test_dst_change(trigger_args, start_date, start_date_fold, correct_next_date):
    """
    Making sure that CronTrigger works correctly when crossing the DST switch threshold.
    Note that you should explicitly compare datetimes as strings to avoid the internal datetime
    comparison which would test for equality in the UTC timezone.

    """
    timezone = ZoneInfo("US/Eastern")
    trigger = CronTrigger(timezone=timezone, **trigger_args)
    start_date = start_date.replace(tzinfo=timezone, fold=start_date_fold)
    correct_next_date = correct_next_date.replace(tzinfo=timezone, fold=1)
    assert str(trigger.get_next_fire_time(None, start_date)) == str(correct_next_date)


def test_timezone_change(timezone):
    """
    Ensure that get_next_fire_time method returns datetimes in the timezone of the trigger and
    not in the timezone of the passed in start_date.

    """
    est = pytz.FixedOffset(-300)
    cst = pytz.FixedOffset(-360)
    trigger = CronTrigger(hour=11, minute="*/5", timezone=est)
    start_date = cst.localize(datetime(2009, 9, 26, 10, 16))
    correct_next_date = est.localize(datetime(2009, 9, 26, 11, 20))
    assert str(trigger.get_next_fire_time(None, start_date)) == str(correct_next_date)


def test_pickle(timezone):
    """Test that the trigger is pickleable."""

    trigger = CronTrigger(
        year=2016,
        month="5-6",
        day="20-28",
        hour=7,
        minute=25,
        second="*",
        timezone=timezone,
    )
    data = pickle.dumps(trigger, 2)
    trigger2 = pickle.loads(data)

    for attr in CronTrigger.__slots__:
        assert getattr(trigger2, attr) == getattr(trigger, attr)


def test_jitter_produces_differrent_valid_results(timezone):
    trigger = CronTrigger(minute="*", jitter=5)
    now = localize(datetime(2017, 11, 12, 6, 55, 30), timezone)

    results = set()
    for _ in range(0, 100):
        next_fire_time = trigger.get_next_fire_time(None, now)
        results.add(next_fire_time)
        assert timedelta(seconds=25) <= (next_fire_time - now) <= timedelta(seconds=35)

    assert 1 < len(results)


def test_jitter_with_timezone(timezone):
    est = pytz.FixedOffset(-300)
    cst = pytz.FixedOffset(-360)
    trigger = CronTrigger(hour=11, minute="*/5", timezone=est, jitter=5)
    start_date = cst.localize(datetime(2009, 9, 26, 10, 16))
    correct_next_date = est.localize(datetime(2009, 9, 26, 11, 20))
    for _ in range(0, 100):
        assert abs(
            trigger.get_next_fire_time(None, start_date) - correct_next_date
        ) <= timedelta(seconds=5)


@pytest.mark.parametrize(
    "trigger_args, start_date, fold, correct_next_date",
    [
        ({"hour": 8}, datetime(2013, 3, 9, 12), False, datetime(2013, 3, 10, 8)),
        ({"hour": 8}, datetime(2013, 11, 2, 12), True, datetime(2013, 11, 3, 8)),
        (
            {"minute": "*/30"},
            datetime(2013, 3, 10, 1, 35),
            False,
            datetime(2013, 3, 10, 3),
        ),
        (
            {"minute": "*/30"},
            datetime(2013, 11, 3, 1, 35),
            True,
            datetime(2013, 11, 3, 1),
        ),
    ],
    ids=[
        "absolute_spring",
        "absolute_autumn",
        "interval_spring",
        "interval_autumn",
    ],
)
def test_jitter_dst_change(trigger_args, start_date, fold, correct_next_date):
    timezone = ZoneInfo("US/Eastern")
    trigger = CronTrigger(timezone=timezone, jitter=5, **trigger_args)
    start_date = start_date.replace(tzinfo=timezone)
    correct_next_date = correct_next_date.replace(tzinfo=timezone)

    for _ in range(0, 100):
        next_fire_time = trigger.get_next_fire_time(None, start_date)
        assert abs(next_fire_time - correct_next_date) <= timedelta(seconds=5)


def test_jitter_with_end_date(timezone):
    now = localize(datetime(2017, 11, 12, 6, 55, 30), timezone)
    end_date = localize(datetime(2017, 11, 12, 6, 56, 0), timezone)
    trigger = CronTrigger(minute="*", jitter=5, end_date=end_date)

    for _ in range(0, 100):
        next_fire_time = trigger.get_next_fire_time(None, now)
        assert next_fire_time is None or next_fire_time <= end_date


@pytest.mark.parametrize(
    "values, expected",
    [
        (
            dict(day="*/31"),
            r"Error validating expression '\*/31': the step value \(31\) is higher "
            r"than the total range of the expression \(30\)",
        ),
        (
            dict(day="4-6/3"),
            r"Error validating expression '4-6/3': the step value \(3\) is higher "
            r"than the total range of the expression \(2\)",
        ),
        (
            dict(hour="0-24"),
            r"Error validating expression '0-24': the last value \(24\) is higher "
            r"than the maximum value \(23\)",
        ),
        (
            dict(day="0-3"),
            r"Error validating expression '0-3': the first value \(0\) is lower "
            r"than the minimum value \(1\)",
        ),
    ],
    ids=[
        "too_large_step_all",
        "too_large_step_range",
        "too_high_last",
        "too_low_first",
    ],
)
def test_invalid_ranges(values, expected):
    pytest.raises(ValueError, CronTrigger, **values).match(expected)


@pytest.mark.parametrize(
    "expr, expected_repr",
    [
        (
            "* * * * *",
            "<CronTrigger (month='*', day='*', day_of_week='*', hour='*', minute='*', "
            "timezone='Europe/Berlin')>",
        ),
        (
            "0-14 * 14-28 jul fri",
            "<CronTrigger (month='jul', day='14-28', day_of_week='fri', hour='*', minute='0-14', "
            "timezone='Europe/Berlin')>",
        ),
        (
            " 0-14   * 14-28   jul       fri",
            "<CronTrigger (month='jul', day='14-28', day_of_week='fri', hour='*', minute='0-14', "
            "timezone='Europe/Berlin')>",
        ),
    ],
    ids=["always", "assorted", "multiple_spaces_in_format"],
)
def test_from_crontab(expr, expected_repr, timezone):
    trigger = CronTrigger.from_crontab(expr, timezone)
    assert repr(trigger) == expected_repr



import sys
from datetime import datetime
from unittest.mock import Mock

import pytest
import pytz

from apscheduler.job import Job
from apscheduler.schedulers.base import BaseScheduler
from apscheduler.schedulers.blocking import BlockingScheduler
from apscheduler.util import localize

if sys.version_info < (3, 9):
    from backports.zoneinfo import ZoneInfo
else:
    from zoneinfo import ZoneInfo


@pytest.fixture
def anyio_backend():
    return "asyncio"


@pytest.fixture(params=["pytz", "zoneinfo"])
def timezone(request, monkeypatch):
    if request.param == "pytz":
        tz = pytz.timezone("Europe/Berlin")
    else:
        tz = ZoneInfo("Europe/Berlin")

    monkeypatch.setattr(
        "apscheduler.schedulers.base.get_localzone", Mock(return_value=tz)
    )
    return tz


@pytest.fixture
def freeze_time(monkeypatch, timezone):
    class TimeFreezer:
        def __init__(self, initial):
            self.current = initial
            self.increment = None

        def get(self, tzinfo=None):
            now = (
                self.current.astimezone(tzinfo)
                if tzinfo
                else self.current.replace(tzinfo=None)
            )
            if self.increment:
                self.current += self.increment
            return now

        def set(self, new_time):
            self.current = new_time

        def next(
            self,
        ):
            return self.current + self.increment

        def set_increment(self, delta):
            self.increment = delta

    freezer = TimeFreezer(localize(datetime(2011, 4, 3, 18, 40), timezone))
    fake_datetime = Mock(datetime, now=freezer.get)
    monkeypatch.setattr("apscheduler.schedulers.base.datetime", fake_datetime)
    monkeypatch.setattr("apscheduler.executors.base.datetime", fake_datetime)
    monkeypatch.setattr("apscheduler.triggers.interval.datetime", fake_datetime)
    monkeypatch.setattr("apscheduler.triggers.date.datetime", fake_datetime)
    return freezer


@pytest.fixture
def job_defaults(timezone):
    run_date = localize(datetime(2011, 4, 3, 18, 40), timezone)
    return {
        "trigger": "date",
        "trigger_args": {"run_date": run_date, "timezone": timezone},
        "executor": "default",
        "args": (),
        "kwargs": {},
        "id": b"t\xc3\xa9st\xc3\xafd".decode("utf-8"),
        "misfire_grace_time": 1,
        "coalesce": False,
        "name": b"n\xc3\xa4m\xc3\xa9".decode("utf-8"),
        "max_instances": 1,
    }


@pytest.fixture
def create_job(job_defaults, timezone):
    def create(**kwargs):
        kwargs.setdefault("scheduler", Mock(BaseScheduler, timezone=timezone))
        job_kwargs = job_defaults.copy()
        job_kwargs.update(kwargs)
        job_kwargs["trigger"] = BlockingScheduler()._create_trigger(
            job_kwargs.pop("trigger"), job_kwargs.pop("trigger_args")
        )
        job_kwargs.setdefault("next_run_time", None)
        return Job(**job_kwargs)

    return create



from datetime import datetime

import pytest

from apscheduler.jobstores.base import ConflictingIdError, JobLookupError
from apscheduler.jobstores.memory import MemoryJobStore
from apscheduler.util import localize


def dummy_job():
    pass


def dummy_job2():
    pass


def dummy_job3():
    pass


class DummyClass:
    def dummy_method(self, a, b):
        return a + b

    @classmethod
    def dummy_classmethod(cls, a, b):
        return a + b


dummy_instance = DummyClass()


@pytest.fixture
def memjobstore():
    yield MemoryJobStore()


@pytest.fixture
def sqlalchemyjobstore(tmpdir):
    db_path = tmpdir.join("apscheduler_unittest.sqlite")
    sqlalchemy = pytest.importorskip("apscheduler.jobstores.sqlalchemy")
    store = sqlalchemy.SQLAlchemyJobStore(url=f"sqlite:///{db_path}")
    store.start(None, "sqlalchemy")
    yield store
    store.shutdown()
    db_path.remove()


@pytest.fixture
def rethinkdbjobstore():
    rethinkdb = pytest.importorskip("apscheduler.jobstores.rethinkdb")
    store = rethinkdb.RethinkDBJobStore(database="apscheduler_unittest")
    store.start(None, "rethinkdb")
    yield store
    store.r.db_drop("apscheduler_unittest").run(store.conn)
    store.shutdown()


@pytest.fixture
def mongodbjobstore():
    mongodb = pytest.importorskip("apscheduler.jobstores.mongodb")
    store = mongodb.MongoDBJobStore(database="apscheduler_unittest")
    store.start(None, "mongodb")
    yield store
    store.client.drop_database(store.collection.database.name)
    store.shutdown()


@pytest.fixture
def redisjobstore():
    redis = pytest.importorskip("apscheduler.jobstores.redis")
    store = redis.RedisJobStore()
    store.start(None, "redis")
    yield store
    store.remove_all_jobs()
    store.shutdown()


@pytest.fixture
def zookeeperjobstore():
    zookeeper = pytest.importorskip("apscheduler.jobstores.zookeeper")
    store = zookeeper.ZooKeeperJobStore(path="/apscheduler_unittest")
    store.start(None, "zookeeper")
    yield store
    store.remove_all_jobs()
    store.shutdown()


@pytest.fixture
def etcdjobstore():
    etcd = pytest.importorskip("apscheduler.jobstores.etcd")
    store = etcd.EtcdJobStore(path="/apscheduler_unittest")
    store.start(None, "etcd")
    yield store
    store.remove_all_jobs()
    store.shutdown()


@pytest.fixture(
    params=[
        "memjobstore",
        "sqlalchemyjobstore",
        "mongodbjobstore",
        "redisjobstore",
        "rethinkdbjobstore",
        "zookeeperjobstore",
        "etcdjobstore",
    ],
    ids=["memory", "sqlalchemy", "mongodb", "redis", "rethinkdb", "zookeeper", "etcd"],
)
def jobstore(request):
    return request.getfixturevalue(request.param)


@pytest.fixture(
    params=[
        "sqlalchemyjobstore",
        "mongodbjobstore",
        "redisjobstore",
        "rethinkdbjobstore",
        "zookeeperjobstore",
        "etcdjobstore",
    ],
    ids=["sqlalchemy", "mongodb", "redis", "rethinkdb", "zookeeper", "etcd"],
)
def persistent_jobstore(request):
    return request.getfixturevalue(request.param)


@pytest.fixture
def create_add_job(timezone, create_job):
    def create(
        jobstore,
        func=dummy_job,
        run_date=datetime(2999, 1, 1),
        id=None,
        paused=False,
        **kwargs,
    ):
        run_date = localize(run_date, timezone)
        job = create_job(
            func=func,
            trigger="date",
            trigger_args={"run_date": run_date},
            id=id,
            **kwargs,
        )
        job.next_run_time = (
            None if paused else job.trigger.get_next_fire_time(None, run_date)
        )
        if jobstore:
            jobstore.add_job(job)
        return job

    return create


def test_add_callable_instance_method_job(jobstore, create_add_job):
    instance = DummyClass()
    initial_job = create_add_job(
        jobstore, instance.dummy_method, kwargs={"a": 1, "b": 2}
    )
    job = jobstore.lookup_job(initial_job.id)
    assert job.func(*job.args, **job.kwargs) == 3


def test_add_callable_class_method_job(jobstore, create_add_job):
    initial_job = create_add_job(
        jobstore, DummyClass.dummy_classmethod, kwargs={"a": 1, "b": 2}
    )
    job = jobstore.lookup_job(initial_job.id)
    assert job.func(*job.args, **job.kwargs) == 3


def test_add_textual_instance_method_job(jobstore, create_add_job):
    initial_job = create_add_job(
        jobstore,
        "tests.test_jobstores:dummy_instance.dummy_method",
        kwargs={"a": 1, "b": 2},
    )
    job = jobstore.lookup_job(initial_job.id)
    assert job.func(*job.args, **job.kwargs) == 3


def test_add_textual_class_method_job(jobstore, create_add_job):
    initial_job = create_add_job(
        jobstore,
        "tests.test_jobstores:DummyClass.dummy_classmethod",
        kwargs={"a": 1, "b": 2},
    )
    job = jobstore.lookup_job(initial_job.id)
    assert job.func(*job.args, **job.kwargs) == 3


def test_lookup_job(jobstore, create_add_job):
    initial_job = create_add_job(jobstore)
    job = jobstore.lookup_job(initial_job.id)
    assert job == initial_job


def test_lookup_nonexistent_job(jobstore):
    assert jobstore.lookup_job("foo") is None


def test_get_all_jobs(jobstore, create_add_job):
    job1 = create_add_job(jobstore, dummy_job, datetime(2016, 5, 3))
    job2 = create_add_job(jobstore, dummy_job2, datetime(2013, 8, 14))
    job3 = create_add_job(jobstore, dummy_job2, datetime(2013, 7, 11), paused=True)
    jobs = jobstore.get_all_jobs()
    assert jobs == [job2, job1, job3]


def test_get_pending_jobs(jobstore, create_add_job, timezone):
    create_add_job(jobstore, dummy_job, datetime(2016, 5, 3))
    job2 = create_add_job(jobstore, dummy_job2, datetime(2014, 2, 26))
    job3 = create_add_job(jobstore, dummy_job3, datetime(2013, 8, 14))
    create_add_job(jobstore, dummy_job3, datetime(2013, 7, 11), paused=True)
    jobs = jobstore.get_due_jobs(localize(datetime(2014, 2, 27), timezone))
    assert jobs == [job3, job2]

    jobs = jobstore.get_due_jobs(localize(datetime(2013, 8, 13), timezone))
    assert jobs == []


def test_get_pending_jobs_subsecond_difference(jobstore, create_add_job, timezone):
    job1 = create_add_job(jobstore, dummy_job, datetime(2014, 7, 7, 0, 0, 0, 401))
    job2 = create_add_job(jobstore, dummy_job2, datetime(2014, 7, 7, 0, 0, 0, 402))
    job3 = create_add_job(jobstore, dummy_job3, datetime(2014, 7, 7, 0, 0, 0, 400))
    jobs = jobstore.get_due_jobs(localize(datetime(2014, 7, 7, 1), timezone))
    assert jobs == [job3, job1, job2]


def test_get_next_run_time(jobstore, create_add_job, timezone):
    create_add_job(jobstore, dummy_job, datetime(2016, 5, 3))
    create_add_job(jobstore, dummy_job2, datetime(2014, 2, 26))
    create_add_job(jobstore, dummy_job3, datetime(2013, 8, 14))
    create_add_job(jobstore, dummy_job3, datetime(2013, 7, 11), paused=True)
    assert jobstore.get_next_run_time() == localize(datetime(2013, 8, 14), timezone)


def test_add_job_conflicting_id(jobstore, create_add_job):
    create_add_job(jobstore, dummy_job, datetime(2016, 5, 3), id="blah")
    pytest.raises(
        ConflictingIdError,
        create_add_job,
        jobstore,
        dummy_job2,
        datetime(2014, 2, 26),
        id="blah",
    )


def test_update_job(jobstore, create_add_job, timezone):
    job1 = create_add_job(jobstore, dummy_job, datetime(2016, 5, 3))
    job2 = create_add_job(jobstore, dummy_job2, datetime(2014, 2, 26))
    replacement = create_add_job(
        None, dummy_job, datetime(2016, 5, 4), id=job1.id, max_instances=6
    )
    assert replacement.max_instances == 6
    jobstore.update_job(replacement)

    jobs = jobstore.get_all_jobs()
    assert len(jobs) == 2
    assert jobs[0].id == job2.id
    assert jobs[1].id == job1.id
    assert jobs[1].next_run_time == localize(datetime(2016, 5, 4), timezone)
    assert jobs[1].max_instances == 6


@pytest.mark.parametrize(
    "next_run_time", [datetime(2013, 8, 13), None], ids=["earlier", "null"]
)
def test_update_job_next_runtime(jobstore, create_add_job, next_run_time, timezone):
    job1 = create_add_job(jobstore, dummy_job, datetime(2016, 5, 3))
    create_add_job(jobstore, dummy_job2, datetime(2014, 2, 26))
    job3 = create_add_job(jobstore, dummy_job3, datetime(2013, 8, 14))
    job1.next_run_time = localize(next_run_time, timezone) if next_run_time else None
    jobstore.update_job(job1)

    if next_run_time:
        assert jobstore.get_next_run_time() == job1.next_run_time
    else:
        assert jobstore.get_next_run_time() == job3.next_run_time


@pytest.mark.parametrize(
    "next_run_time", [datetime(2013, 8, 13), None], ids=["earlier", "null"]
)
@pytest.mark.parametrize("index", [0, 1, 2], ids=["first", "middle", "last"])
def test_update_job_clear_next_runtime(
    jobstore, create_add_job, next_run_time, timezone, index
):
    """
    Tests that update_job() maintains the proper ordering of the jobs,
    even when their next run times are initially the same.

    """
    jobs = [
        create_add_job(jobstore, dummy_job, datetime(2014, 2, 26), "job%d" % i)
        for i in range(3)
    ]
    jobs[index].next_run_time = (
        localize(next_run_time, timezone) if next_run_time else None
    )
    jobstore.update_job(jobs[index])
    due_date = localize(datetime(2014, 2, 27), timezone)
    due_jobs = jobstore.get_due_jobs(due_date)

    assert len(due_jobs) == (3 if next_run_time else 2)
    due_job_ids = [job.id for job in due_jobs]
    if next_run_time:
        if index == 0:
            assert due_job_ids == ["job0", "job1", "job2"]
        elif index == 1:
            assert due_job_ids == ["job1", "job0", "job2"]
        else:
            assert due_job_ids == ["job2", "job0", "job1"]
    else:
        if index == 0:
            assert due_job_ids == ["job1", "job2"]
        elif index == 1:
            assert due_job_ids == ["job0", "job2"]
        else:
            assert due_job_ids == ["job0", "job1"]


def test_update_job_nonexistent_job(jobstore, create_add_job):
    job = create_add_job(None, dummy_job, datetime(2016, 5, 3))
    pytest.raises(JobLookupError, jobstore.update_job, job)


def test_one_job_fails_to_load(
    persistent_jobstore, create_add_job, monkeypatch, timezone
):
    job1 = create_add_job(persistent_jobstore, dummy_job, datetime(2016, 5, 3))
    job2 = create_add_job(persistent_jobstore, dummy_job2, datetime(2014, 2, 26))
    create_add_job(persistent_jobstore, dummy_job3, datetime(2013, 8, 14))

    # Make the dummy_job2 function disappear
    monkeypatch.delitem(globals(), "dummy_job3")

    jobs = persistent_jobstore.get_all_jobs()
    assert jobs == [job2, job1]

    assert persistent_jobstore.get_next_run_time() == localize(
        datetime(2014, 2, 26), timezone
    )


def test_remove_job(jobstore, create_add_job):
    job1 = create_add_job(jobstore, dummy_job, datetime(2016, 5, 3))
    job2 = create_add_job(jobstore, dummy_job2, datetime(2014, 2, 26))

    jobstore.remove_job(job1.id)
    jobs = jobstore.get_all_jobs()
    assert jobs == [job2]

    jobstore.remove_job(job2.id)
    jobs = jobstore.get_all_jobs()
    assert jobs == []


def test_remove_nonexistent_job(jobstore):
    pytest.raises(JobLookupError, jobstore.remove_job, "blah")


def test_remove_all_jobs(jobstore, create_add_job):
    create_add_job(jobstore, dummy_job, datetime(2016, 5, 3))
    create_add_job(jobstore, dummy_job2, datetime(2014, 2, 26))

    jobstore.remove_all_jobs()
    jobs = jobstore.get_all_jobs()
    assert jobs == []


def test_repr_memjobstore(memjobstore):
    assert repr(memjobstore) == "<MemoryJobStore>"


def test_repr_sqlalchemyjobstore(sqlalchemyjobstore):
    assert repr(sqlalchemyjobstore).startswith("<SQLAlchemyJobStore (url=")


def test_repr_mongodbjobstore(mongodbjobstore):
    assert repr(mongodbjobstore).startswith("<MongoDBJobStore (client=MongoClient(")


def test_repr_redisjobstore(redisjobstore):
    assert repr(redisjobstore) == "<RedisJobStore>"


def test_repr_zookeeperjobstore(zookeeperjobstore):
    class_sig = "<ZooKeeperJobStore (client=<kazoo.client.KazooClient"
    assert repr(zookeeperjobstore).startswith(class_sig)


def test_repr_etcdjobstore(etcdjobstore):
    class_sig = "<EtcdJobStore (client=<etcd3.client"
    assert repr(etcdjobstore).startswith(class_sig)


def test_memstore_close(memjobstore, create_add_job):
    create_add_job(memjobstore, dummy_job, datetime(2016, 5, 3))
    memjobstore.shutdown()
    assert not memjobstore.get_all_jobs()


def test_sqlalchemy_engine_ref():
    global sqla_engine
    sqlalchemy = pytest.importorskip("apscheduler.jobstores.sqlalchemy")
    sqla_engine = sqlalchemy.create_engine("sqlite:///")
    try:
        sqlalchemy.SQLAlchemyJobStore(engine=f"{__name__}:sqla_engine")
    finally:
        sqla_engine.dispose()
        del sqla_engine


def test_sqlalchemy_missing_engine():
    sqlalchemy = pytest.importorskip("apscheduler.jobstores.sqlalchemy")
    exc = pytest.raises(ValueError, sqlalchemy.SQLAlchemyJobStore)
    assert "Need either" in str(exc.value)


def test_mongodb_client_ref():
    global mongodb_client
    mongodb = pytest.importorskip("apscheduler.jobstores.mongodb")
    mongodb_client = mongodb.MongoClient()
    try:
        mongodb.MongoDBJobStore(client=f"{__name__}:mongodb_client")
    finally:
        mongodb_client.close()
        del mongodb_client


def test_zookeeper_client_ref():
    global zookeeper_client
    zookeeper = pytest.importorskip("apscheduler.jobstores.zookeeper")
    zookeeper_client = zookeeper.KazooClient()
    try:
        zookeeperjobstore = zookeeper.ZooKeeperJobStore(
            client=f"{__name__}:zookeeper_client"
        )
        zookeeperjobstore.start(None, "zookeeper")
        zookeeperjobstore.shutdown()
        assert zookeeper_client.connected is True
    finally:
        zookeeper_client.stop()
        zookeeper_client.close()
        del zookeeper_client


def test_zookeeper_client_keep_open():
    global zookeeper_client
    zookeeper = pytest.importorskip("apscheduler.jobstores.zookeeper")
    zookeeper_client = zookeeper.KazooClient()
    try:
        zookeeperjobstore = zookeeper.ZooKeeperJobStore(
            client=f"{__name__}:zookeeper_client", close_connection_on_exit=True
        )
        zookeeperjobstore.start(None, "zookeeper")
        zookeeperjobstore.shutdown()
        assert zookeeper_client.connected is False
    finally:
        del zookeeper_client


def test_mongodb_null_database():
    mongodb = pytest.importorskip("apscheduler.jobstores.mongodb")
    exc = pytest.raises(ValueError, mongodb.MongoDBJobStore, database="")
    assert '"database"' in str(exc.value)


def test_mongodb_null_collection():
    mongodb = pytest.importorskip("apscheduler.jobstores.mongodb")
    exc = pytest.raises(ValueError, mongodb.MongoDBJobStore, collection="")
    assert '"collection"' in str(exc.value)


def test_zookeeper_null_path():
    zookeeper = pytest.importorskip("apscheduler.jobstores.zookeeper")
    exc = pytest.raises(ValueError, zookeeper.ZooKeeperJobStore, path="")
    assert '"path"' in str(exc.value)


def test_etcd_null_path():
    etcd = pytest.importorskip("apscheduler.jobstores.etcd")
    exc = pytest.raises(ValueError, etcd.EtcdJobStore, path="")
    assert '"path"' in str(exc.value)


def test_etcd_client_ref():
    global etcd_client
    etcd = pytest.importorskip("apscheduler.jobstores.etcd")
    etcd_client = etcd.Etcd3Client()
    try:
        etcdjobstore = etcd.EtcdJobStore(client=f"{__name__}:etcd_client")
        etcdjobstore.start(None, "etcd")
        etcdjobstore.shutdown()
    finally:
        etcd_client.close()
        del etcd_client



import importlib.metadata as importlib_metadata
import sys

try:
    release = importlib_metadata.version("APScheduler").split("-")[0]
except importlib_metadata.PackageNotFoundError:
    release = "3.5.0"

version_info = tuple(int(x) if x.isdigit() else x for x in release.split("."))
version = __version__ = ".".join(str(x) for x in version_info[:3])
del sys, importlib_metadata



from collections.abc import Iterable, Mapping
from inspect import isclass, ismethod
from uuid import uuid4

from apscheduler.triggers.base import BaseTrigger
from apscheduler.util import (
    check_callable_args,
    convert_to_datetime,
    datetime_repr,
    get_callable_name,
    obj_to_ref,
    ref_to_obj,
)


class Job:
    """
    Contains the options given when scheduling callables and its current schedule and other state.
    This class should never be instantiated by the user.

    :var str id: the unique identifier of this job
    :var str name: the description of this job
    :var func: the callable to execute
    :var tuple|list args: positional arguments to the callable
    :var dict kwargs: keyword arguments to the callable
    :var bool coalesce: whether to only run the job once when several run times are due
    :var trigger: the trigger object that controls the schedule of this job
    :var str executor: the name of the executor that will run this job
    :var int misfire_grace_time: the time (in seconds) how much this job's execution is allowed to
        be late (``None`` means "allow the job to run no matter how late it is")
    :var int max_instances: the maximum number of concurrently executing instances allowed for this
        job
    :var datetime.datetime next_run_time: the next scheduled run time of this job

    .. note::
        The ``misfire_grace_time`` has some non-obvious effects on job execution. See the
        :ref:`missed-job-executions` section in the documentation for an in-depth explanation.
    """

    __slots__ = (
        "_scheduler",
        "_jobstore_alias",
        "id",
        "trigger",
        "executor",
        "func",
        "func_ref",
        "args",
        "kwargs",
        "name",
        "misfire_grace_time",
        "coalesce",
        "max_instances",
        "next_run_time",
        "__weakref__",
    )

    def __init__(self, scheduler, id=None, **kwargs):
        super().__init__()
        self._scheduler = scheduler
        self._jobstore_alias = None
        self._modify(id=id or uuid4().hex, **kwargs)

    def modify(self, **changes):
        """
        Makes the given changes to this job and saves it in the associated job store.

        Accepted keyword arguments are the same as the variables on this class.

        .. seealso:: :meth:`~apscheduler.schedulers.base.BaseScheduler.modify_job`

        :return Job: this job instance

        """
        self._scheduler.modify_job(self.id, self._jobstore_alias, **changes)
        return self

    def reschedule(self, trigger, **trigger_args):
        """
        Shortcut for switching the trigger on this job.

        .. seealso:: :meth:`~apscheduler.schedulers.base.BaseScheduler.reschedule_job`

        :return Job: this job instance

        """
        self._scheduler.reschedule_job(
            self.id, self._jobstore_alias, trigger, **trigger_args
        )
        return self

    def pause(self):
        """
        Temporarily suspend the execution of this job.

        .. seealso:: :meth:`~apscheduler.schedulers.base.BaseScheduler.pause_job`

        :return Job: this job instance

        """
        self._scheduler.pause_job(self.id, self._jobstore_alias)
        return self

    def resume(self):
        """
        Resume the schedule of this job if previously paused.

        .. seealso:: :meth:`~apscheduler.schedulers.base.BaseScheduler.resume_job`

        :return Job: this job instance

        """
        self._scheduler.resume_job(self.id, self._jobstore_alias)
        return self

    def remove(self):
        """
        Unschedules this job and removes it from its associated job store.

        .. seealso:: :meth:`~apscheduler.schedulers.base.BaseScheduler.remove_job`

        """
        self._scheduler.remove_job(self.id, self._jobstore_alias)

    @property
    def pending(self):
        """
        Returns ``True`` if the referenced job is still waiting to be added to its designated job
        store.

        """
        return self._jobstore_alias is None

    #
    # Private API
    #

    def _get_run_times(self, now):
        """
        Computes the scheduled run times between ``next_run_time`` and ``now`` (inclusive).

        :type now: datetime.datetime
        :rtype: list[datetime.datetime]

        """
        run_times = []
        next_run_time = self.next_run_time
        while next_run_time and next_run_time <= now:
            run_times.append(next_run_time)
            next_run_time = self.trigger.get_next_fire_time(next_run_time, now)

        return run_times

    def _modify(self, **changes):
        """
        Validates the changes to the Job and makes the modifications if and only if all of them
        validate.

        """
        approved = {}

        if "id" in changes:
            value = changes.pop("id")
            if not isinstance(value, str):
                raise TypeError("id must be a nonempty string")
            if hasattr(self, "id"):
                raise ValueError("The job ID may not be changed")
            approved["id"] = value

        if "func" in changes or "args" in changes or "kwargs" in changes:
            func = changes.pop("func") if "func" in changes else self.func
            args = changes.pop("args") if "args" in changes else self.args
            kwargs = changes.pop("kwargs") if "kwargs" in changes else self.kwargs

            if isinstance(func, str):
                func_ref = func
                func = ref_to_obj(func)
            elif callable(func):
                try:
                    func_ref = obj_to_ref(func)
                except ValueError:
                    # If this happens, this Job won't be serializable
                    func_ref = None
            else:
                raise TypeError("func must be a callable or a textual reference to one")

            if not hasattr(self, "name") and changes.get("name", None) is None:
                changes["name"] = get_callable_name(func)

            if isinstance(args, str) or not isinstance(args, Iterable):
                raise TypeError("args must be a non-string iterable")
            if isinstance(kwargs, str) or not isinstance(kwargs, Mapping):
                raise TypeError("kwargs must be a dict-like object")

            check_callable_args(func, args, kwargs)

            approved["func"] = func
            approved["func_ref"] = func_ref
            approved["args"] = args
            approved["kwargs"] = kwargs

        if "name" in changes:
            value = changes.pop("name")
            if not value or not isinstance(value, str):
                raise TypeError("name must be a nonempty string")
            approved["name"] = value

        if "misfire_grace_time" in changes:
            value = changes.pop("misfire_grace_time")
            if value is not None and (not isinstance(value, int) or value <= 0):
                raise TypeError(
                    "misfire_grace_time must be either None or a positive integer"
                )
            approved["misfire_grace_time"] = value

        if "coalesce" in changes:
            value = bool(changes.pop("coalesce"))
            approved["coalesce"] = value

        if "max_instances" in changes:
            value = changes.pop("max_instances")
            if not isinstance(value, int) or value <= 0:
                raise TypeError("max_instances must be a positive integer")
            approved["max_instances"] = value

        if "trigger" in changes:
            trigger = changes.pop("trigger")
            if not isinstance(trigger, BaseTrigger):
                raise TypeError(
                    f"Expected a trigger instance, got {trigger.__class__.__name__} instead"
                )

            approved["trigger"] = trigger

        if "executor" in changes:
            value = changes.pop("executor")
            if not isinstance(value, str):
                raise TypeError("executor must be a string")
            approved["executor"] = value

        if "next_run_time" in changes:
            value = changes.pop("next_run_time")
            approved["next_run_time"] = convert_to_datetime(
                value, self._scheduler.timezone, "next_run_time"
            )

        if changes:
            raise AttributeError(
                "The following are not modifiable attributes of Job: {}".format(
                    ", ".join(changes)
                )
            )

        for key, value in approved.items():
            setattr(self, key, value)

    def __getstate__(self):
        # Don't allow this Job to be serialized if the function reference could not be determined
        if not self.func_ref:
            raise ValueError(
                f"This Job cannot be serialized since the reference to its callable ({self.func!r}) could not "
                "be determined. Consider giving a textual reference (module:function name) "
                "instead."
            )

        # Instance methods cannot survive serialization as-is, so store the "self" argument
        # explicitly
        func = self.func
        if (
            ismethod(func)
            and not isclass(func.__self__)
            and obj_to_ref(func) == self.func_ref
        ):
            args = (func.__self__,) + tuple(self.args)
        else:
            args = self.args

        return {
            "version": 1,
            "id": self.id,
            "func": self.func_ref,
            "trigger": self.trigger,
            "executor": self.executor,
            "args": args,
            "kwargs": self.kwargs,
            "name": self.name,
            "misfire_grace_time": self.misfire_grace_time,
            "coalesce": self.coalesce,
            "max_instances": self.max_instances,
            "next_run_time": self.next_run_time,
        }

    def __setstate__(self, state):
        if state.get("version", 1) > 1:
            raise ValueError(
                f"Job has version {state['version']}, but only version 1 can be handled"
            )

        self.id = state["id"]
        self.func_ref = state["func"]
        self.func = ref_to_obj(self.func_ref)
        self.trigger = state["trigger"]
        self.executor = state["executor"]
        self.args = state["args"]
        self.kwargs = state["kwargs"]
        self.name = state["name"]
        self.misfire_grace_time = state["misfire_grace_time"]
        self.coalesce = state["coalesce"]
        self.max_instances = state["max_instances"]
        self.next_run_time = state["next_run_time"]

    def __eq__(self, other):
        if isinstance(other, Job):
            return self.id == other.id
        return NotImplemented

    def __repr__(self):
        return f"<Job (id={self.id} name={self.name})>"

    def __str__(self):
        if hasattr(self, "next_run_time"):
            status = (
                "next run at: " + datetime_repr(self.next_run_time)
                if self.next_run_time
                else "paused"
            )
        else:
            status = "pending"

        return f"{self.name} (trigger: {self.trigger}, {status})"



import pickle
from datetime import datetime, timedelta

import pytest

from apscheduler.triggers.combining import AndTrigger, BaseCombiningTrigger, OrTrigger
from apscheduler.triggers.cron import CronTrigger
from apscheduler.util import localize


class TestAndTrigger:
    @pytest.fixture
    def trigger(self, timezone):
        return AndTrigger(
            [
                CronTrigger(
                    month="5-8",
                    day="6-15",
                    end_date=localize(datetime(2017, 8, 10), timezone),
                ),
                CronTrigger(
                    month="6-9",
                    day="*/3",
                    end_date=localize(datetime(2017, 9, 7), timezone),
                ),
            ]
        )

    @pytest.mark.parametrize(
        "start_time, expected",
        [
            (datetime(2017, 8, 6), datetime(2017, 8, 7)),
            (datetime(2017, 8, 10, 1), None),
        ],
        ids=["firstmatch", "end"],
    )
    def test_next_fire_time(self, trigger, timezone, start_time, expected):
        expected = localize(expected, timezone) if expected else None
        assert (
            trigger.get_next_fire_time(None, localize(start_time, timezone)) == expected
        )

    def test_jitter(self, trigger, timezone):
        trigger.jitter = 5
        start_time = localize(datetime(2017, 8, 6), timezone)
        expected = localize(datetime(2017, 8, 7), timezone)
        for _ in range(100):
            next_fire_time = trigger.get_next_fire_time(None, start_time)
            assert abs(expected - next_fire_time) <= timedelta(seconds=5)

    @pytest.mark.parametrize("jitter", [None, 5], ids=["nojitter", "jitter"])
    def test_repr(self, trigger, jitter):
        trigger.jitter = jitter
        jitter_part = f", jitter={jitter}" if jitter else ""
        assert repr(trigger) == (
            "<AndTrigger([<CronTrigger (month='5-8', day='6-15', "
            "end_date='2017-08-10 00:00:00 CEST', timezone='Europe/Berlin')>, <CronTrigger "
            "(month='6-9', day='*/3', end_date='2017-09-07 00:00:00 CEST', "
            f"timezone='Europe/Berlin')>]{jitter_part})>"
        )

    def test_str(self, trigger):
        assert (
            str(trigger)
            == "and[cron[month='5-8', day='6-15'], cron[month='6-9', day='*/3']]"
        )

    @pytest.mark.parametrize("jitter", [None, 5], ids=["nojitter", "jitter"])
    def test_pickle(self, trigger, jitter):
        """Test that the trigger is pickleable."""
        trigger.jitter = jitter
        data = pickle.dumps(trigger, 2)
        trigger2 = pickle.loads(data)

        for attr in BaseCombiningTrigger.__slots__:
            assert repr(getattr(trigger2, attr)) == repr(getattr(trigger, attr))


class TestOrTrigger:
    @pytest.fixture
    def trigger(self, timezone):
        return OrTrigger(
            [
                CronTrigger(
                    month="5-8",
                    day="6-15",
                    end_date=localize(datetime(2017, 8, 10), timezone),
                ),
                CronTrigger(
                    month="6-9",
                    day="*/3",
                    end_date=localize(datetime(2017, 9, 7), timezone),
                ),
            ]
        )

    @pytest.mark.parametrize(
        "start_time, expected",
        [(datetime(2017, 8, 6), datetime(2017, 8, 6)), (datetime(2017, 9, 7, 1), None)],
        ids=["earliest", "end"],
    )
    def test_next_fire_time(self, trigger, timezone, start_time, expected):
        expected = localize(expected, timezone) if expected else None
        assert (
            trigger.get_next_fire_time(None, localize(start_time, timezone)) == expected
        )

    def test_jitter(self, trigger, timezone):
        trigger.jitter = 5
        start_time = expected = localize(datetime(2017, 8, 6), timezone)
        for _ in range(100):
            next_fire_time = trigger.get_next_fire_time(None, start_time)
            assert abs(expected - next_fire_time) <= timedelta(seconds=5)

    @pytest.mark.parametrize("jitter", [None, 5], ids=["nojitter", "jitter"])
    def test_repr(self, trigger, jitter):
        trigger.jitter = jitter
        jitter_part = f", jitter={jitter}" if jitter else ""
        assert repr(trigger) == (
            "<OrTrigger([<CronTrigger (month='5-8', day='6-15', "
            "end_date='2017-08-10 00:00:00 CEST', timezone='Europe/Berlin')>, <CronTrigger "
            "(month='6-9', day='*/3', end_date='2017-09-07 00:00:00 CEST', "
            f"timezone='Europe/Berlin')>]{jitter_part})>"
        )

    def test_str(self, trigger):
        assert (
            str(trigger)
            == "or[cron[month='5-8', day='6-15'], cron[month='6-9', day='*/3']]"
        )

    @pytest.mark.parametrize("jitter", [None, 5], ids=["nojitter", "jitter"])
    def test_pickle(self, trigger, jitter):
        """Test that the trigger is pickleable."""
        trigger.jitter = jitter
        data = pickle.dumps(trigger, 2)
        trigger2 = pickle.loads(data)

        for attr in BaseCombiningTrigger.__slots__:
            assert repr(getattr(trigger2, attr)) == repr(getattr(trigger, attr))



"""
Demonstrates how to use the Tornado compatible scheduler to schedule a job that executes on 3
second intervals.
"""

import os
from datetime import datetime

from tornado.ioloop import IOLoop

from apscheduler.schedulers.tornado import TornadoScheduler


def tick():
    print(f"Tick! The time is: {datetime.now()}")


if __name__ == "__main__":
    scheduler = TornadoScheduler()
    scheduler.add_job(tick, "interval", seconds=3)
    scheduler.start()
    print("Press Ctrl+{} to exit".format("Break" if os.name == "nt" else "C"))

    # Execution will block here until Ctrl+C (Ctrl+Break on Windows) is pressed.
    try:
        IOLoop.instance().start()
    except (KeyboardInterrupt, SystemExit):
        pass



import pickle
import warnings

from apscheduler.job import Job
from apscheduler.jobstores.base import BaseJobStore, ConflictingIdError, JobLookupError
from apscheduler.util import (
    datetime_to_utc_timestamp,
    maybe_ref,
    utc_timestamp_to_datetime,
)

try:
    from bson.binary import Binary
    from pymongo import ASCENDING, MongoClient
    from pymongo.errors import DuplicateKeyError
except ImportError as exc:  # pragma: nocover
    raise ImportError("MongoDBJobStore requires PyMongo installed") from exc


class MongoDBJobStore(BaseJobStore):
    """
    Stores jobs in a MongoDB database. Any leftover keyword arguments are directly passed to
    pymongo's `MongoClient
    <http://api.mongodb.org/python/current/api/pymongo/mongo_client.html#pymongo.mongo_client.MongoClient>`_.

    Plugin alias: ``mongodb``

    :param str database: database to store jobs in
    :param str collection: collection to store jobs in
    :param client: a :class:`~pymongo.mongo_client.MongoClient` instance to use instead of
        providing connection arguments
    :param int pickle_protocol: pickle protocol level to use (for serialization), defaults to the
        highest available
    """

    def __init__(
        self,
        database="apscheduler",
        collection="jobs",
        client=None,
        pickle_protocol=pickle.HIGHEST_PROTOCOL,
        **connect_args,
    ):
        super().__init__()
        self.pickle_protocol = pickle_protocol

        if not database:
            raise ValueError('The "database" parameter must not be empty')
        if not collection:
            raise ValueError('The "collection" parameter must not be empty')

        if client:
            self.client = maybe_ref(client)
        else:
            connect_args.setdefault("w", 1)
            self.client = MongoClient(**connect_args)

        self.collection = self.client[database][collection]

    def start(self, scheduler, alias):
        super().start(scheduler, alias)
        self.collection.create_index("next_run_time", sparse=True)

    @property
    def connection(self):
        warnings.warn(
            'The "connection" member is deprecated -- use "client" instead',
            DeprecationWarning,
        )
        return self.client

    def lookup_job(self, job_id):
        document = self.collection.find_one(job_id, ["job_state"])
        return self._reconstitute_job(document["job_state"]) if document else None

    def get_due_jobs(self, now):
        timestamp = datetime_to_utc_timestamp(now)
        return self._get_jobs({"next_run_time": {"$lte": timestamp}})

    def get_next_run_time(self):
        document = self.collection.find_one(
            {"next_run_time": {"$ne": None}},
            projection=["next_run_time"],
            sort=[("next_run_time", ASCENDING)],
        )
        return (
            utc_timestamp_to_datetime(document["next_run_time"]) if document else None
        )

    def get_all_jobs(self):
        jobs = self._get_jobs({})
        self._fix_paused_jobs_sorting(jobs)
        return jobs

    def add_job(self, job):
        try:
            self.collection.insert_one(
                {
                    "_id": job.id,
                    "next_run_time": datetime_to_utc_timestamp(job.next_run_time),
                    "job_state": Binary(
                        pickle.dumps(job.__getstate__(), self.pickle_protocol)
                    ),
                }
            )
        except DuplicateKeyError:
            raise ConflictingIdError(job.id)

    def update_job(self, job):
        changes = {
            "next_run_time": datetime_to_utc_timestamp(job.next_run_time),
            "job_state": Binary(pickle.dumps(job.__getstate__(), self.pickle_protocol)),
        }
        result = self.collection.update_one({"_id": job.id}, {"$set": changes})
        if result and result.matched_count == 0:
            raise JobLookupError(job.id)

    def remove_job(self, job_id):
        result = self.collection.delete_one({"_id": job_id})
        if result and result.deleted_count == 0:
            raise JobLookupError(job_id)

    def remove_all_jobs(self):
        self.collection.delete_many({})

    def shutdown(self):
        self.client.close()

    def _reconstitute_job(self, job_state):
        job_state = pickle.loads(job_state)
        job = Job.__new__(Job)
        job.__setstate__(job_state)
        job._scheduler = self._scheduler
        job._jobstore_alias = self._alias
        return job

    def _get_jobs(self, conditions):
        jobs = []
        failed_job_ids = []
        for document in self.collection.find(
            conditions, ["_id", "job_state"], sort=[("next_run_time", ASCENDING)]
        ):
            try:
                jobs.append(self._reconstitute_job(document["job_state"]))
            except BaseException:
                self._logger.exception(
                    'Unable to restore job "%s" -- removing it', document["_id"]
                )
                failed_job_ids.append(document["_id"])

        # Remove all the jobs we failed to restore
        if failed_job_ids:
            self.collection.delete_many({"_id": {"$in": failed_job_ids}})

        return jobs

    def __repr__(self):
        return f"<{self.__class__.__name__} (client={self.client})>"



import gc
import os
import signal
import time
from asyncio import CancelledError
from datetime import datetime
from threading import Event
from types import TracebackType
from unittest.mock import MagicMock, Mock, patch

import pytest
from pytz import UTC, utc

from apscheduler.events import EVENT_JOB_ERROR, EVENT_JOB_EXECUTED, EVENT_JOB_MISSED
from apscheduler.executors.asyncio import AsyncIOExecutor
from apscheduler.executors.base import (
    MaxInstancesReachedError,
    run_coroutine_job,
    run_job,
)
from apscheduler.executors.pool import ProcessPoolExecutor
from apscheduler.executors.tornado import TornadoExecutor
from apscheduler.job import Job
from apscheduler.schedulers.asyncio import AsyncIOScheduler
from apscheduler.schedulers.background import BackgroundScheduler
from apscheduler.schedulers.base import BaseScheduler
from apscheduler.schedulers.tornado import TornadoScheduler
from apscheduler.util import localize


@pytest.fixture
def mock_scheduler(timezone):
    scheduler_ = Mock(BaseScheduler, timezone=timezone)
    scheduler_._create_lock = MagicMock()
    return scheduler_


@pytest.fixture(params=["threadpool", "processpool"])
def executor(request, mock_scheduler):
    if request.param == "threadpool":
        from apscheduler.executors.pool import ThreadPoolExecutor

        executor_ = ThreadPoolExecutor()
    else:
        from apscheduler.executors.pool import ProcessPoolExecutor

        executor_ = ProcessPoolExecutor()

    executor_.start(mock_scheduler, "dummy")
    yield executor_
    executor_.shutdown()


@pytest.fixture
async def asyncio_scheduler():
    scheduler = AsyncIOScheduler()
    scheduler.start(paused=True)
    yield scheduler
    scheduler.shutdown(False)


@pytest.fixture
async def asyncio_executor(asyncio_scheduler):
    executor = AsyncIOExecutor()
    executor.start(asyncio_scheduler, "default")
    yield executor
    executor.shutdown()


@pytest.fixture
def tornado_scheduler():
    scheduler = TornadoScheduler()
    scheduler.start(paused=True)
    yield scheduler
    scheduler.shutdown(False)


@pytest.fixture
def tornado_executor(tornado_scheduler):
    executor = TornadoExecutor()
    executor.start(tornado_scheduler, "default")
    yield executor
    executor.shutdown()


async def waiter(sleep, exception):
    await sleep(0.1)
    if exception:
        raise Exception("dummy error")
    else:
        return True


def wait_event():
    time.sleep(0.2)
    return "test"


def failure():
    raise Exception("test failure")


def success():
    return 5


def test_max_instances(mock_scheduler, executor, create_job, timezone):
    """Tests that the maximum instance limit on a job is respected."""
    events = []
    mock_scheduler._dispatch_event = lambda event: events.append(event)
    now = datetime.now(timezone)
    job = create_job(func=wait_event, max_instances=2, next_run_time=None)
    executor.submit_job(job, [now])
    executor.submit_job(job, [now])

    pytest.raises(MaxInstancesReachedError, executor.submit_job, job, [now])
    executor.shutdown()
    assert len(events) == 2
    assert events[0].retval == "test"
    assert events[1].retval == "test"


@pytest.mark.parametrize(
    "event_code,func",
    [
        (EVENT_JOB_EXECUTED, success),
        (EVENT_JOB_MISSED, failure),
        (EVENT_JOB_ERROR, failure),
    ],
    ids=["executed", "missed", "error"],
)
def test_submit_job(mock_scheduler, executor, create_job, timezone, event_code, func):
    """
    Tests that an EVENT_JOB_EXECUTED event is delivered to the scheduler if the job was
    successfully executed.

    """
    mock_scheduler._dispatch_event = MagicMock()
    job = create_job(func=func, id="foo")
    job._jobstore_alias = "test_jobstore"
    now = datetime.now(timezone)
    run_time = (
        localize(datetime(1970, 1, 1), timezone)
        if event_code == EVENT_JOB_MISSED
        else now
    )
    executor.submit_job(job, [run_time])
    executor.shutdown()

    assert mock_scheduler._dispatch_event.call_count == 1
    event = mock_scheduler._dispatch_event.call_args[0][0]
    assert event.code == event_code
    assert event.job_id == "foo"
    assert event.jobstore == "test_jobstore"

    if event_code == EVENT_JOB_EXECUTED:
        assert event.retval == 5
    elif event_code == EVENT_JOB_ERROR:
        assert str(event.exception) == "test failure"
        assert isinstance(event.traceback, str)


class FauxJob:
    id = "abc"
    max_instances = 1
    _jobstore_alias = "foo"


def dummy_run_job(job, jobstore_alias, run_times, logger_name):
    raise Exception("dummy")


def test_run_job_error(monkeypatch, executor):
    """Tests that _run_job_error is properly called if an exception is raised in run_job()"""

    def run_job_error(job_id, exc, traceback):
        assert job_id == "abc"
        exc_traceback[:] = [exc, traceback]
        event.set()

    event = Event()
    exc_traceback = [None, None]
    monkeypatch.setattr("apscheduler.executors.base.run_job", dummy_run_job)
    monkeypatch.setattr("apscheduler.executors.pool.run_job", dummy_run_job)
    monkeypatch.setattr(executor, "_run_job_error", run_job_error)
    executor.submit_job(FauxJob(), [])

    event.wait(5)
    assert str(exc_traceback[0]) == "dummy"
    if exc_traceback[1] is not None:
        assert isinstance(exc_traceback[1], TracebackType)


def test_run_job_memory_leak():
    class FooBar:
        pass

    def func():
        foo = FooBar()  # noqa: F841
        raise Exception("dummy")

    fake_job = Mock(Job, func=func, args=(), kwargs={}, misfire_grace_time=1)
    with patch("logging.getLogger"):
        for _ in range(5):
            run_job(fake_job, "foo", [datetime.now(UTC)], __name__)

    foos = [x for x in gc.get_objects() if type(x) is FooBar]
    assert len(foos) == 0


@pytest.mark.anyio
async def test_run_async_job_memory_leak():
    class FooBar:
        pass

    async def func():
        foo = FooBar()  # noqa: F841
        raise Exception("dummy")

    fake_job = Mock(Job, func=func, args=(), kwargs={}, misfire_grace_time=1)
    with patch("logging.getLogger"):
        for _ in range(5):
            await run_coroutine_job(fake_job, "foo", [datetime.now(utc)], __name__)

    foos = [x for x in gc.get_objects() if type(x) is FooBar]
    assert len(foos) == 0


def test_broken_pool():
    def listener(evt):
        pid[0] = evt.retval
        event.set()

    pid = [None]
    event = Event()
    scheduler = BackgroundScheduler(executors={"default": ProcessPoolExecutor(1)})
    scheduler.add_listener(listener, EVENT_JOB_EXECUTED)
    scheduler.add_job(os.getpid, "date", run_date=datetime.now(UTC))
    scheduler.start()

    event.wait(3)
    killed_pid = pid[0]
    os.kill(pid[0], signal.SIGTERM)
    try:
        os.waitpid(pid[0], 0)
    except OSError:
        pass

    event.clear()
    scheduler.add_job(os.getpid, "date", run_date=datetime.now(UTC))
    event.wait(3)
    assert pid[0] != killed_pid
    scheduler.shutdown(True)


@pytest.mark.parametrize("exception", [False, True])
@pytest.mark.anyio
async def test_run_coroutine_job(asyncio_scheduler, asyncio_executor, exception):
    from asyncio import Future, sleep

    future = Future()
    job = asyncio_scheduler.add_job(
        waiter, "interval", seconds=1, args=[sleep, exception]
    )
    asyncio_executor._run_job_success = lambda job_id, events: future.set_result(events)
    asyncio_executor._run_job_error = lambda job_id, exc, tb: future.set_exception(exc)
    asyncio_executor.submit_job(job, [datetime.now(utc)])
    events = await future
    assert len(events) == 1
    if exception:
        assert str(events[0].exception) == "dummy error"
    else:
        assert events[0].retval is True


@pytest.mark.parametrize("exception", [False, True])
@pytest.mark.anyio
async def test_run_coroutine_job_tornado(
    tornado_scheduler, tornado_executor, exception
):
    from tornado.concurrent import Future
    from tornado.gen import sleep

    future = Future()
    job = tornado_scheduler.add_job(
        waiter, "interval", seconds=1, args=[sleep, exception]
    )
    tornado_executor._run_job_success = lambda job_id, events: future.set_result(events)
    tornado_executor._run_job_error = lambda job_id, exc, tb: future.set_exception(exc)
    tornado_executor.submit_job(job, [datetime.now(utc)])
    events = await future
    assert len(events) == 1
    if exception:
        assert str(events[0].exception) == "dummy error"
    else:
        assert events[0].retval is True


@pytest.mark.anyio
async def test_asyncio_executor_shutdown(asyncio_scheduler, asyncio_executor):
    """Test that the AsyncIO executor cancels its pending tasks on shutdown."""
    from asyncio import sleep

    job = asyncio_scheduler.add_job(waiter, "interval", seconds=1, args=[sleep, None])
    asyncio_executor.submit_job(job, [datetime.now(utc)])
    futures = asyncio_executor._pending_futures.copy()
    assert len(futures) == 1

    asyncio_executor.shutdown()
    with pytest.raises(CancelledError):
        await futures.pop()



import pickle
import sys
from datetime import date, datetime, timedelta
from unittest.mock import Mock

import pytest
import pytz

from apscheduler.triggers.cron import CronTrigger
from apscheduler.triggers.interval import IntervalTrigger
from apscheduler.util import localize

if sys.version_info < (3, 9):
    from backports.zoneinfo import ZoneInfo
else:
    from zoneinfo import ZoneInfo


@pytest.fixture()
def trigger(timezone):
    return IntervalTrigger(
        seconds=1, start_date=datetime(2009, 8, 4, second=2), timezone=timezone
    )


def test_invalid_interval(timezone):
    pytest.raises(TypeError, IntervalTrigger, "1-6", timezone=timezone)


def test_start_end_times_string(timezone, monkeypatch):
    monkeypatch.setattr(
        "apscheduler.triggers.interval.get_localzone", Mock(return_value=timezone)
    )
    trigger = IntervalTrigger(
        start_date="2016-11-05 05:06:53", end_date="2017-11-05 05:11:32"
    )
    assert trigger.start_date == localize(datetime(2016, 11, 5, 5, 6, 53), timezone)
    assert trigger.end_date == localize(datetime(2017, 11, 5, 5, 11, 32), timezone)


def test_before(trigger, timezone):
    """Tests that if "start_date" is later than "now", it will return start_date."""
    now = trigger.start_date - timedelta(seconds=2)
    assert trigger.get_next_fire_time(None, now) == trigger.start_date


def test_within(trigger, timezone):
    """
    Tests that if "now" is between "start_date" and the next interval, it will return the next
    interval.

    """
    now = trigger.start_date + timedelta(microseconds=1000)
    assert (
        trigger.get_next_fire_time(None, now) == trigger.start_date + trigger.interval
    )


def test_no_start_date(timezone):
    trigger = IntervalTrigger(seconds=2, timezone=timezone)
    now = datetime.now(timezone)
    assert (trigger.get_next_fire_time(None, now) - now) <= timedelta(seconds=2)


def test_different_tz(trigger, timezone):
    alter_tz = pytz.FixedOffset(-60)
    start_date = alter_tz.localize(datetime(2009, 8, 3, 22, second=2, microsecond=1000))
    correct_next_date = localize(datetime(2009, 8, 4, 1, second=3), timezone)
    assert trigger.get_next_fire_time(None, start_date) == correct_next_date


def test_end_date(timezone):
    """Tests that the interval trigger won't return any datetimes past the set end time."""
    start_date = localize(datetime(2014, 5, 26), timezone)
    trigger = IntervalTrigger(
        minutes=5,
        start_date=start_date,
        end_date=datetime(2014, 5, 26, 0, 7),
        timezone=timezone,
    )
    assert trigger.get_next_fire_time(
        None, start_date + timedelta(minutes=2)
    ) == start_date.replace(minute=5)
    assert trigger.get_next_fire_time(None, start_date + timedelta(minutes=6)) is None


def test_dst_change():
    """
    Making sure that IntervalTrigger works during the ambiguous "fall-back" DST period.
    Note that you should explicitly compare datetimes as strings to avoid the internal datetime
    comparison which would test for equality in the UTC timezone.

    """
    eastern = ZoneInfo("US/Eastern")
    start_date = datetime(2013, 3, 1)  # Start within EDT
    trigger = IntervalTrigger(hours=1, start_date=start_date, timezone=eastern)

    datetime_edt = datetime(2013, 3, 10, 1, 5, tzinfo=eastern)
    correct_next_date = datetime(2013, 3, 10, 3, tzinfo=eastern)
    assert str(trigger.get_next_fire_time(None, datetime_edt)) == str(correct_next_date)

    datetime_est = datetime(2013, 11, 3, 1, 5, tzinfo=eastern)
    correct_next_date = datetime(2013, 11, 3, 1, tzinfo=eastern, fold=1)
    next_fire_time = trigger.get_next_fire_time(None, datetime_est)
    assert next_fire_time == correct_next_date
    assert str(next_fire_time) == str(correct_next_date)


def test_space_in_expr(timezone):
    trigger = CronTrigger(day="1-2, 4-7", timezone=timezone)
    assert repr(trigger) == "<CronTrigger (day='1-2,4-7', timezone='Europe/Berlin')>"


def test_repr(trigger):
    assert repr(trigger) == (
        "<IntervalTrigger (interval=datetime.timedelta(seconds=1), "
        "start_date='2009-08-04 00:00:02 CEST', "
        "timezone='Europe/Berlin')>"
    )


def test_str(trigger):
    assert str(trigger) == "interval[0:00:01]"


def test_pickle(timezone):
    """Test that the trigger is pickleable."""

    trigger = IntervalTrigger(
        weeks=2,
        days=6,
        minutes=13,
        seconds=2,
        start_date=date(2016, 4, 3),
        timezone=timezone,
        jitter=12,
    )
    data = pickle.dumps(trigger, 2)
    trigger2 = pickle.loads(data)

    for attr in IntervalTrigger.__slots__:
        assert getattr(trigger2, attr) == getattr(trigger, attr)


def test_jitter_produces_different_valid_results(timezone):
    trigger = IntervalTrigger(seconds=5, timezone=timezone, jitter=3)
    now = datetime.now(timezone)

    results = set()
    for _ in range(0, 100):
        next_fire_time = trigger.get_next_fire_time(None, now)
        results.add(next_fire_time)
        assert timedelta(seconds=2) <= (next_fire_time - now) <= timedelta(seconds=8)
    assert 1 < len(results)


@pytest.mark.parametrize(
    "trigger_args, start_date, start_date_dst, correct_next_date",
    [
        (
            {"hours": 1},
            datetime(2013, 3, 10, 1, 35),
            False,
            datetime(2013, 3, 10, 3, 35),
        ),
        (
            {"hours": 1},
            datetime(2013, 11, 3, 1, 35),
            True,
            datetime(2013, 11, 3, 1, 35),
        ),
    ],
    ids=["interval_spring", "interval_autumn"],
)
def test_jitter_dst_change(trigger_args, start_date, start_date_dst, correct_next_date):
    timezone = pytz.timezone("US/Eastern")
    epsilon = timedelta(seconds=1)
    start_date = timezone.localize(start_date, is_dst=start_date_dst)
    trigger = IntervalTrigger(
        timezone=timezone, start_date=start_date, jitter=5, **trigger_args
    )
    correct_next_date = timezone.localize(correct_next_date, is_dst=not start_date_dst)

    for _ in range(0, 100):
        next_fire_time = trigger.get_next_fire_time(None, start_date + epsilon)
        assert abs(next_fire_time - correct_next_date) <= timedelta(seconds=5)


def test_jitter_with_end_date(timezone):
    now = localize(datetime(2017, 11, 12, 6, 55, 58), timezone)
    end_date = localize(datetime(2017, 11, 12, 6, 56, 0), timezone)
    trigger = IntervalTrigger(seconds=5, jitter=5, end_date=end_date)

    for _ in range(0, 100):
        next_fire_time = trigger.get_next_fire_time(None, now)
        assert next_fire_time is None or next_fire_time <= end_date



"""
This example demonstrates the use of the Redis job store.
On each run, it adds a new alarm that fires after ten seconds.
You can exit the program, restart it and observe that any previous alarms that have not fired yet
are still active. Running the example with the --clear switch will remove any existing alarms.
"""

import os
import sys
from datetime import datetime, timedelta

from apscheduler.schedulers.blocking import BlockingScheduler


def alarm(time):
    print(f"Alarm! This alarm was scheduled at {time}.")


if __name__ == "__main__":
    scheduler = BlockingScheduler()
    scheduler.add_jobstore(
        "redis", jobs_key="example.jobs", run_times_key="example.run_times"
    )
    if len(sys.argv) > 1 and sys.argv[1] == "--clear":
        scheduler.remove_all_jobs()

    alarm_time = datetime.now() + timedelta(seconds=10)
    scheduler.add_job(alarm, "date", run_date=alarm_time, args=[datetime.now()])
    print("To clear the alarms, run this example with the --clear argument.")
    print("Press Ctrl+{} to exit".format("Break" if os.name == "nt" else "C"))

    try:
        scheduler.start()
    except (KeyboardInterrupt, SystemExit):
        pass



"""
Demonstrates how to use the Qt compatible scheduler to schedule a job that executes on 3 second
intervals.
"""

import signal
import sys
from datetime import datetime
from importlib import import_module
from itertools import product

from apscheduler.schedulers.qt import QtScheduler

for version, pkgname in product(range(6, 1, -1), ("PySide", "PyQt")):
    try:
        qtwidgets = import_module(pkgname + str(version) + ".QtWidgets")
    except ImportError:
        pass
    else:
        QApplication = qtwidgets.QApplication
        QLabel = qtwidgets.QLabel
        break
else:
    raise ImportError(
        "Could not import the QtWidgets module from either PySide or PyQt"
    )


def tick():
    label.setText(f"Tick! The time is: {datetime.now()}")


if __name__ == "__main__":
    app = QApplication(sys.argv)

    # This enables processing of Ctrl+C keypresses
    signal.signal(signal.SIGINT, lambda *args: QApplication.quit())

    label = QLabel("The timer text will appear here in a moment!")
    label.setWindowTitle("QtScheduler example")
    label.setFixedSize(280, 50)
    label.show()

    scheduler = QtScheduler()
    scheduler.add_job(tick, "interval", seconds=3)
    scheduler.start()

    # Execution will block here until the user closes the windows or Ctrl+C is pressed.
    app.exec()



"""
Demonstrates how to schedule a job to be run in a process pool on 3 second intervals.
"""

import os
from datetime import datetime

from apscheduler.schedulers.blocking import BlockingScheduler


def tick():
    print(f"Tick! The time is: {datetime.now()}")


if __name__ == "__main__":
    scheduler = BlockingScheduler()
    scheduler.add_executor("processpool")
    scheduler.add_job(tick, "interval", seconds=3)
    print("Press Ctrl+{} to exit".format("Break" if os.name == "nt" else "C"))

    try:
        scheduler.start()
    except (KeyboardInterrupt, SystemExit):
        pass



"""
Basic example showing how to schedule a callable using a textual reference.
"""

import os

from apscheduler.schedulers.blocking import BlockingScheduler

if __name__ == "__main__":
    scheduler = BlockingScheduler()
    scheduler.add_job("sys:stdout.write", "interval", seconds=3, args=["tick\n"])
    print("Press Ctrl+{} to exit".format("Break" if os.name == "nt" else "C"))

    try:
        scheduler.start()
    except (KeyboardInterrupt, SystemExit):
        pass



import pickle
from datetime import date, datetime, timedelta

import pytest
import pytz

from apscheduler.triggers.date import DateTrigger
from apscheduler.util import localize


@pytest.mark.parametrize(
    "run_date,alter_tz,previous,now,expected",
    [
        (
            datetime(2009, 7, 6),
            None,
            None,
            datetime(2008, 5, 4),
            datetime(2009, 7, 6),
        ),
        (
            datetime(2009, 7, 6),
            None,
            None,
            datetime(2009, 7, 6),
            datetime(2009, 7, 6),
        ),
        (
            datetime(2009, 7, 6),
            None,
            None,
            datetime(2009, 9, 2),
            datetime(2009, 7, 6),
        ),
        ("2009-7-6", None, None, datetime(2009, 9, 2), datetime(2009, 7, 6)),
        (
            datetime(2009, 7, 6),
            None,
            datetime(2009, 7, 6),
            datetime(2009, 9, 2),
            None,
        ),
        (
            datetime(2009, 7, 5, 22),
            pytz.FixedOffset(-60),
            datetime(2009, 7, 6),
            datetime(2009, 7, 6),
            None,
        ),
        (
            None,
            pytz.FixedOffset(-120),
            None,
            datetime(2011, 4, 3, 18, 40),
            datetime(2011, 4, 3, 18, 40),
        ),
    ],
    ids=[
        "earlier",
        "exact",
        "later",
        "as text",
        "previously fired",
        "alternate timezone",
        "current_time",
    ],
)
def test_get_next_fire_time(
    run_date, alter_tz, previous, now, expected, timezone, freeze_time
):
    trigger = DateTrigger(run_date, alter_tz or timezone)
    previous = localize(previous, timezone) if previous else None
    now = localize(now, timezone)
    expected = localize(expected, timezone) if expected else None
    assert trigger.get_next_fire_time(previous, now) == expected


@pytest.mark.parametrize(
    "is_dst", [True, False], ids=["daylight saving", "standard time"]
)
def test_dst_change(is_dst):
    """
    Test that DateTrigger works during the ambiguous "fall-back" DST period.

    Note that you should explicitly compare datetimes as strings to avoid the internal datetime
    comparison which would test for equality in the UTC timezone.

    """
    eastern = pytz.timezone("US/Eastern")
    run_date = eastern.localize(datetime(2013, 10, 3, 1, 5), is_dst=is_dst)

    fire_date = eastern.normalize(run_date + timedelta(minutes=55))
    trigger = DateTrigger(run_date=fire_date, timezone=eastern)
    assert str(trigger.get_next_fire_time(None, fire_date)) == str(fire_date)


def test_repr(timezone):
    trigger = DateTrigger(datetime(2009, 7, 6), timezone)
    assert repr(trigger) == "<DateTrigger (run_date='2009-07-06 00:00:00 CEST')>"


def test_str(timezone):
    trigger = DateTrigger(datetime(2009, 7, 6), timezone)
    assert str(trigger) == "date[2009-07-06 00:00:00 CEST]"


def test_pickle(timezone):
    """Test that the trigger is pickleable."""
    trigger = DateTrigger(date(2016, 4, 3), timezone=timezone)
    data = pickle.dumps(trigger, 2)
    trigger2 = pickle.loads(data)
    assert trigger2.run_date == trigger.run_date



import concurrent.futures
import multiprocessing
from abc import abstractmethod
from concurrent.futures.process import BrokenProcessPool

from apscheduler.executors.base import BaseExecutor, run_job


class BasePoolExecutor(BaseExecutor):
    @abstractmethod
    def __init__(self, pool):
        super().__init__()
        self._pool = pool

    def _do_submit_job(self, job, run_times):
        def callback(f):
            exc, tb = (
                f.exception_info()
                if hasattr(f, "exception_info")
                else (f.exception(), getattr(f.exception(), "__traceback__", None))
            )
            if exc:
                self._run_job_error(job.id, exc, tb)
            else:
                self._run_job_success(job.id, f.result())

        f = self._pool.submit(
            run_job, job, job._jobstore_alias, run_times, self._logger.name
        )
        f.add_done_callback(callback)

    def shutdown(self, wait=True):
        self._pool.shutdown(wait)


class ThreadPoolExecutor(BasePoolExecutor):
    """
    An executor that runs jobs in a concurrent.futures thread pool.

    Plugin alias: ``threadpool``

    :param max_workers: the maximum number of spawned threads.
    :param pool_kwargs: dict of keyword arguments to pass to the underlying
        ThreadPoolExecutor constructor
    """

    def __init__(self, max_workers=10, pool_kwargs=None):
        pool_kwargs = pool_kwargs or {}
        pool = concurrent.futures.ThreadPoolExecutor(int(max_workers), **pool_kwargs)
        super().__init__(pool)


class ProcessPoolExecutor(BasePoolExecutor):
    """
    An executor that runs jobs in a concurrent.futures process pool.

    Plugin alias: ``processpool``

    :param max_workers: the maximum number of spawned processes.
    :param pool_kwargs: dict of keyword arguments to pass to the underlying
        ProcessPoolExecutor constructor
    """

    def __init__(self, max_workers=10, pool_kwargs=None):
        self.pool_kwargs = pool_kwargs or {}
        self.pool_kwargs.setdefault("mp_context", multiprocessing.get_context("spawn"))
        pool = concurrent.futures.ProcessPoolExecutor(
            int(max_workers), **self.pool_kwargs
        )
        super().__init__(pool)

    def _do_submit_job(self, job, run_times):
        try:
            super()._do_submit_job(job, run_times)
        except BrokenProcessPool:
            self._logger.warning(
                "Process pool is broken; replacing pool with a fresh instance"
            )
            self._pool = self._pool.__class__(
                self._pool._max_workers, **self.pool_kwargs
            )
            super()._do_submit_job(job, run_times)



import random
from datetime import datetime, timedelta
from math import ceil

from tzlocal import get_localzone

from apscheduler.triggers.base import BaseTrigger
from apscheduler.util import (
    astimezone,
    convert_to_datetime,
    datetime_repr,
)


class IntervalTrigger(BaseTrigger):
    """
    Triggers on specified intervals, starting on ``start_date`` if specified, ``datetime.now()`` +
    interval otherwise.

    :param int weeks: number of weeks to wait
    :param int days: number of days to wait
    :param int hours: number of hours to wait
    :param int minutes: number of minutes to wait
    :param int seconds: number of seconds to wait
    :param datetime|str start_date: starting point for the interval calculation
    :param datetime|str end_date: latest possible date/time to trigger on
    :param datetime.tzinfo|str timezone: time zone to use for the date/time calculations
    :param int|None jitter: delay the job execution by ``jitter`` seconds at most
    """

    __slots__ = (
        "timezone",
        "start_date",
        "end_date",
        "interval",
        "interval_length",
        "jitter",
    )

    def __init__(
        self,
        weeks=0,
        days=0,
        hours=0,
        minutes=0,
        seconds=0,
        start_date=None,
        end_date=None,
        timezone=None,
        jitter=None,
    ):
        self.interval = timedelta(
            weeks=weeks, days=days, hours=hours, minutes=minutes, seconds=seconds
        )
        self.interval_length = self.interval.total_seconds()
        if self.interval_length == 0:
            self.interval = timedelta(seconds=1)
            self.interval_length = 1

        if timezone:
            self.timezone = astimezone(timezone)
        elif isinstance(start_date, datetime) and start_date.tzinfo:
            self.timezone = astimezone(start_date.tzinfo)
        elif isinstance(end_date, datetime) and end_date.tzinfo:
            self.timezone = astimezone(end_date.tzinfo)
        else:
            self.timezone = get_localzone()

        start_date = start_date or (datetime.now(self.timezone) + self.interval)
        self.start_date = convert_to_datetime(start_date, self.timezone, "start_date")
        self.end_date = convert_to_datetime(end_date, self.timezone, "end_date")

        self.jitter = jitter

    def get_next_fire_time(self, previous_fire_time, now):
        if previous_fire_time:
            next_fire_time = previous_fire_time.timestamp() + self.interval_length
        elif self.start_date > now:
            next_fire_time = self.start_date.timestamp()
        else:
            timediff = now.timestamp() - self.start_date.timestamp()
            next_interval_num = int(ceil(timediff / self.interval_length))
            next_fire_time = (
                self.start_date.timestamp() + self.interval_length * next_interval_num
            )

        if self.jitter is not None:
            next_fire_time += random.uniform(0, self.jitter)

        if not self.end_date or next_fire_time <= self.end_date.timestamp():
            return datetime.fromtimestamp(next_fire_time, tz=self.timezone)

    def __getstate__(self):
        return {
            "version": 2,
            "timezone": astimezone(self.timezone),
            "start_date": self.start_date,
            "end_date": self.end_date,
            "interval": self.interval,
            "jitter": self.jitter,
        }

    def __setstate__(self, state):
        # This is for compatibility with APScheduler 3.0.x
        if isinstance(state, tuple):
            state = state[1]

        if state.get("version", 1) > 2:
            raise ValueError(
                f"Got serialized data for version {state['version']} of "
                f"{self.__class__.__name__}, but only versions up to 2 can be handled"
            )

        self.timezone = state["timezone"]
        self.start_date = state["start_date"]
        self.end_date = state["end_date"]
        self.interval = state["interval"]
        self.interval_length = self.interval.total_seconds()
        self.jitter = state.get("jitter")

    def __str__(self):
        return f"interval[{self.interval!s}]"

    def __repr__(self):
        options = [
            f"interval={self.interval!r}",
            f"start_date={datetime_repr(self.start_date)!r}",
        ]
        if self.end_date:
            options.append(f"end_date={datetime_repr(self.end_date)!r}")
        if self.jitter:
            options.append(f"jitter={self.jitter}")

        return "<{} ({}, timezone='{}')>".format(
            self.__class__.__name__,
            ", ".join(options),
            self.timezone,
        )



"""This module contains the expressions applicable for CronTrigger's fields."""

__all__ = (
    "AllExpression",
    "RangeExpression",
    "WeekdayRangeExpression",
    "WeekdayPositionExpression",
    "LastDayOfMonthExpression",
)

import re
from calendar import monthrange

from apscheduler.util import asint

WEEKDAYS = ["mon", "tue", "wed", "thu", "fri", "sat", "sun"]
MONTHS = [
    "jan",
    "feb",
    "mar",
    "apr",
    "may",
    "jun",
    "jul",
    "aug",
    "sep",
    "oct",
    "nov",
    "dec",
]


class AllExpression:
    value_re = re.compile(r"\*(?:/(?P<step>\d+))?$")

    def __init__(self, step=None):
        self.step = asint(step)
        if self.step == 0:
            raise ValueError("Increment must be higher than 0")

    def validate_range(self, field_name):
        from apscheduler.triggers.cron.fields import MAX_VALUES, MIN_VALUES

        value_range = MAX_VALUES[field_name] - MIN_VALUES[field_name]
        if self.step and self.step > value_range:
            raise ValueError(
                f"the step value ({self.step}) is higher than the total range of the "
                f"expression ({value_range})"
            )

    def get_next_value(self, date, field):
        start = field.get_value(date)
        minval = field.get_min(date)
        maxval = field.get_max(date)
        start = max(start, minval)

        if not self.step:
            next = start
        else:
            distance_to_next = (self.step - (start - minval)) % self.step
            next = start + distance_to_next

        if next <= maxval:
            return next

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.step == other.step

    def __str__(self):
        if self.step:
            return "*/%d" % self.step
        return "*"

    def __repr__(self):
        return f"{self.__class__.__name__}({self.step})"


class RangeExpression(AllExpression):
    value_re = re.compile(r"(?P<first>\d+)(?:-(?P<last>\d+))?(?:/(?P<step>\d+))?$")

    def __init__(self, first, last=None, step=None):
        super().__init__(step)
        first = asint(first)
        last = asint(last)
        if last is None and step is None:
            last = first
        if last is not None and first > last:
            raise ValueError(
                "The minimum value in a range must not be higher than the maximum"
            )
        self.first = first
        self.last = last

    def validate_range(self, field_name):
        from apscheduler.triggers.cron.fields import MAX_VALUES, MIN_VALUES

        super().validate_range(field_name)
        if self.first < MIN_VALUES[field_name]:
            raise ValueError(
                f"the first value ({self.first}) is lower than the minimum value ({MIN_VALUES[field_name]})"
            )
        if self.last is not None and self.last > MAX_VALUES[field_name]:
            raise ValueError(
                f"the last value ({self.last}) is higher than the maximum value ({MAX_VALUES[field_name]})"
            )
        value_range = (self.last or MAX_VALUES[field_name]) - self.first
        if self.step and self.step > value_range:
            raise ValueError(
                f"the step value ({self.step}) is higher than the total range of the "
                f"expression ({value_range})"
            )

    def get_next_value(self, date, field):
        startval = field.get_value(date)
        minval = field.get_min(date)
        maxval = field.get_max(date)

        # Apply range limits
        minval = max(minval, self.first)
        maxval = min(maxval, self.last) if self.last is not None else maxval
        nextval = max(minval, startval)

        # Apply the step if defined
        if self.step:
            distance_to_next = (self.step - (nextval - minval)) % self.step
            nextval += distance_to_next

        return nextval if nextval <= maxval else None

    def __eq__(self, other):
        return (
            isinstance(other, self.__class__)
            and self.first == other.first
            and self.last == other.last
        )

    def __str__(self):
        if self.last != self.first and self.last is not None:
            range = "%d-%d" % (self.first, self.last)
        else:
            range = str(self.first)

        if self.step:
            return "%s/%d" % (range, self.step)

        return range

    def __repr__(self):
        args = [str(self.first)]
        if self.last != self.first and self.last is not None or self.step:
            args.append(str(self.last))

        if self.step:
            args.append(str(self.step))

        return "{}({})".format(self.__class__.__name__, ", ".join(args))


class MonthRangeExpression(RangeExpression):
    value_re = re.compile(r"(?P<first>[a-z]+)(?:-(?P<last>[a-z]+))?", re.IGNORECASE)

    def __init__(self, first, last=None):
        try:
            first_num = MONTHS.index(first.lower()) + 1
        except ValueError:
            raise ValueError(f'Invalid month name "{first}"')

        if last:
            try:
                last_num = MONTHS.index(last.lower()) + 1
            except ValueError:
                raise ValueError(f'Invalid month name "{last}"')
        else:
            last_num = None

        super().__init__(first_num, last_num)

    def __str__(self):
        if self.last != self.first and self.last is not None:
            return f"{MONTHS[self.first - 1]}-{MONTHS[self.last - 1]}"
        return MONTHS[self.first - 1]

    def __repr__(self):
        args = [f"'{MONTHS[self.first]}'"]
        if self.last != self.first and self.last is not None:
            args.append(f"'{MONTHS[self.last - 1]}'")
        return "{}({})".format(self.__class__.__name__, ", ".join(args))


class WeekdayRangeExpression(RangeExpression):
    value_re = re.compile(r"(?P<first>[a-z]+)(?:-(?P<last>[a-z]+))?", re.IGNORECASE)

    def __init__(self, first, last=None):
        try:
            first_num = WEEKDAYS.index(first.lower())
        except ValueError:
            raise ValueError(f'Invalid weekday name "{first}"')

        if last:
            try:
                last_num = WEEKDAYS.index(last.lower())
            except ValueError:
                raise ValueError(f'Invalid weekday name "{last}"')
        else:
            last_num = None

        super().__init__(first_num, last_num)

    def __str__(self):
        if self.last != self.first and self.last is not None:
            return f"{WEEKDAYS[self.first]}-{WEEKDAYS[self.last]}"
        return WEEKDAYS[self.first]

    def __repr__(self):
        args = [f"'{WEEKDAYS[self.first]}'"]
        if self.last != self.first and self.last is not None:
            args.append(f"'{WEEKDAYS[self.last]}'")
        return "{}({})".format(self.__class__.__name__, ", ".join(args))


class WeekdayPositionExpression(AllExpression):
    options = ["1st", "2nd", "3rd", "4th", "5th", "last"]
    value_re = re.compile(
        r"(?P<option_name>{}) +(?P<weekday_name>(?:\d+|\w+))".format("|".join(options)),
        re.IGNORECASE,
    )

    def __init__(self, option_name, weekday_name):
        super().__init__(None)
        try:
            self.option_num = self.options.index(option_name.lower())
        except ValueError:
            raise ValueError(f'Invalid weekday position "{option_name}"')

        try:
            self.weekday = WEEKDAYS.index(weekday_name.lower())
        except ValueError:
            raise ValueError(f'Invalid weekday name "{weekday_name}"')

    def get_next_value(self, date, field):
        # Figure out the weekday of the month's first day and the number of days in that month
        first_day_wday, last_day = monthrange(date.year, date.month)

        # Calculate which day of the month is the first of the target weekdays
        first_hit_day = self.weekday - first_day_wday + 1
        if first_hit_day <= 0:
            first_hit_day += 7

        # Calculate what day of the month the target weekday would be
        if self.option_num < 5:
            target_day = first_hit_day + self.option_num * 7
        else:
            target_day = first_hit_day + ((last_day - first_hit_day) // 7) * 7

        if target_day <= last_day and target_day >= date.day:
            return target_day

    def __eq__(self, other):
        return (
            super().__eq__(other)
            and self.option_num == other.option_num
            and self.weekday == other.weekday
        )

    def __str__(self):
        return f"{self.options[self.option_num]} {WEEKDAYS[self.weekday]}"

    def __repr__(self):
        return f"{self.__class__.__name__}('{self.options[self.option_num]}', '{WEEKDAYS[self.weekday]}')"


class LastDayOfMonthExpression(AllExpression):
    value_re = re.compile(r"last", re.IGNORECASE)

    def __init__(self):
        super().__init__(None)

    def get_next_value(self, date, field):
        return monthrange(date.year, date.month)[1]

    def __str__(self):
        return "last"

    def __repr__(self):
        return f"{self.__class__.__name__}()"



__all__ = (
    "EVENT_SCHEDULER_STARTED",
    "EVENT_SCHEDULER_SHUTDOWN",
    "EVENT_SCHEDULER_PAUSED",
    "EVENT_SCHEDULER_RESUMED",
    "EVENT_EXECUTOR_ADDED",
    "EVENT_EXECUTOR_REMOVED",
    "EVENT_JOBSTORE_ADDED",
    "EVENT_JOBSTORE_REMOVED",
    "EVENT_ALL_JOBS_REMOVED",
    "EVENT_JOB_ADDED",
    "EVENT_JOB_REMOVED",
    "EVENT_JOB_MODIFIED",
    "EVENT_JOB_EXECUTED",
    "EVENT_JOB_ERROR",
    "EVENT_JOB_MISSED",
    "EVENT_JOB_SUBMITTED",
    "EVENT_JOB_MAX_INSTANCES",
    "EVENT_ALL",
    "SchedulerEvent",
    "JobEvent",
    "JobExecutionEvent",
    "JobSubmissionEvent",
)


EVENT_SCHEDULER_STARTED = EVENT_SCHEDULER_START = 2**0
EVENT_SCHEDULER_SHUTDOWN = 2**1
EVENT_SCHEDULER_PAUSED = 2**2
EVENT_SCHEDULER_RESUMED = 2**3
EVENT_EXECUTOR_ADDED = 2**4
EVENT_EXECUTOR_REMOVED = 2**5
EVENT_JOBSTORE_ADDED = 2**6
EVENT_JOBSTORE_REMOVED = 2**7
EVENT_ALL_JOBS_REMOVED = 2**8
EVENT_JOB_ADDED = 2**9
EVENT_JOB_REMOVED = 2**10
EVENT_JOB_MODIFIED = 2**11
EVENT_JOB_EXECUTED = 2**12
EVENT_JOB_ERROR = 2**13
EVENT_JOB_MISSED = 2**14
EVENT_JOB_SUBMITTED = 2**15
EVENT_JOB_MAX_INSTANCES = 2**16
EVENT_ALL = (
    EVENT_SCHEDULER_STARTED
    | EVENT_SCHEDULER_SHUTDOWN
    | EVENT_SCHEDULER_PAUSED
    | EVENT_SCHEDULER_RESUMED
    | EVENT_EXECUTOR_ADDED
    | EVENT_EXECUTOR_REMOVED
    | EVENT_JOBSTORE_ADDED
    | EVENT_JOBSTORE_REMOVED
    | EVENT_ALL_JOBS_REMOVED
    | EVENT_JOB_ADDED
    | EVENT_JOB_REMOVED
    | EVENT_JOB_MODIFIED
    | EVENT_JOB_EXECUTED
    | EVENT_JOB_ERROR
    | EVENT_JOB_MISSED
    | EVENT_JOB_SUBMITTED
    | EVENT_JOB_MAX_INSTANCES
)


class SchedulerEvent:
    """
    An event that concerns the scheduler itself.

    :ivar code: the type code of this event
    :ivar alias: alias of the job store or executor that was added or removed (if applicable)
    """

    def __init__(self, code, alias=None):
        super().__init__()
        self.code = code
        self.alias = alias

    def __repr__(self):
        return "<%s (code=%d)>" % (self.__class__.__name__, self.code)


class JobEvent(SchedulerEvent):
    """
    An event that concerns a job.

    :ivar code: the type code of this event
    :ivar job_id: identifier of the job in question
    :ivar jobstore: alias of the job store containing the job in question
    """

    def __init__(self, code, job_id, jobstore):
        super().__init__(code)
        self.code = code
        self.job_id = job_id
        self.jobstore = jobstore


class JobSubmissionEvent(JobEvent):
    """
    An event that concerns the submission of a job to its executor.

    :ivar scheduled_run_times: a list of datetimes when the job was intended to run
    """

    def __init__(self, code, job_id, jobstore, scheduled_run_times):
        super().__init__(code, job_id, jobstore)
        self.scheduled_run_times = scheduled_run_times


class JobExecutionEvent(JobEvent):
    """
    An event that concerns the running of a job within its executor.

    :ivar scheduled_run_time: the time when the job was scheduled to be run
    :ivar retval: the return value of the successfully executed job
    :ivar exception: the exception raised by the job
    :ivar traceback: a formatted traceback for the exception
    """

    def __init__(
        self,
        code,
        job_id,
        jobstore,
        scheduled_run_time,
        retval=None,
        exception=None,
        traceback=None,
    ):
        super().__init__(code, job_id, jobstore)
        self.scheduled_run_time = scheduled_run_time
        self.retval = retval
        self.exception = exception
        self.traceback = traceback



import pickle
from datetime import datetime, timezone

from apscheduler.job import Job
from apscheduler.jobstores.base import BaseJobStore, ConflictingIdError, JobLookupError
from apscheduler.util import (
    datetime_to_utc_timestamp,
    maybe_ref,
    utc_timestamp_to_datetime,
)

try:
    from etcd3 import Etcd3Client
except ImportError as exc:  # pragma: nocover
    raise ImportError("EtcdJobStore requires etcd3 be installed") from exc


class EtcdJobStore(BaseJobStore):
    """
    Stores jobs in a etcd. Any leftover keyword arguments are directly passed to
    etcd3's `etcd3.client
    <https://python-etcd3.readthedocs.io/en/latest/readme.html>`_.

    Plugin alias: ``etcd``

    :param str path: path to store jobs in
    :param client: a :class:`~etcd3.client.etcd3` instance to use instead of
        providing connection arguments
    :param int pickle_protocol: pickle protocol level to use (for serialization), defaults to the
        highest available
    """

    def __init__(
        self,
        path="/apscheduler",
        client=None,
        close_connection_on_exit=False,
        pickle_protocol=pickle.DEFAULT_PROTOCOL,
        **connect_args,
    ):
        super().__init__()
        self.pickle_protocol = pickle_protocol
        self.close_connection_on_exit = close_connection_on_exit

        if not path:
            raise ValueError('The "path" parameter must not be empty')

        self.path = path

        if client:
            self.client = maybe_ref(client)
        else:
            self.client = Etcd3Client(**connect_args)

    def lookup_job(self, job_id):
        node_path = self.path + "/" + str(job_id)
        try:
            content, _ = self.client.get(node_path)
            content = pickle.loads(content)
            job = self._reconstitute_job(content["job_state"])
            return job
        except BaseException:
            return None

    def get_due_jobs(self, now):
        timestamp = datetime_to_utc_timestamp(now)
        jobs = [
            job_record["job"]
            for job_record in self._get_jobs()
            if job_record["next_run_time"] is not None
            and job_record["next_run_time"] <= timestamp
        ]
        return jobs

    def get_next_run_time(self):
        next_runs = [
            job_record["next_run_time"]
            for job_record in self._get_jobs()
            if job_record["next_run_time"] is not None
        ]
        return utc_timestamp_to_datetime(min(next_runs)) if len(next_runs) > 0 else None

    def get_all_jobs(self):
        jobs = [job_record["job"] for job_record in self._get_jobs()]
        self._fix_paused_jobs_sorting(jobs)
        return jobs

    def add_job(self, job):
        node_path = self.path + "/" + str(job.id)
        value = {
            "next_run_time": datetime_to_utc_timestamp(job.next_run_time),
            "job_state": job.__getstate__(),
        }
        data = pickle.dumps(value, self.pickle_protocol)
        status = self.client.put_if_not_exists(node_path, value=data)
        if not status:
            raise ConflictingIdError(job.id)

    def update_job(self, job):
        node_path = self.path + "/" + str(job.id)
        changes = {
            "next_run_time": datetime_to_utc_timestamp(job.next_run_time),
            "job_state": job.__getstate__(),
        }
        data = pickle.dumps(changes, self.pickle_protocol)
        status, _ = self.client.transaction(
            compare=[self.client.transactions.version(node_path) > 0],
            success=[self.client.transactions.put(node_path, value=data)],
            failure=[],
        )
        if not status:
            raise JobLookupError(job.id)

    def remove_job(self, job_id):
        node_path = self.path + "/" + str(job_id)
        status, _ = self.client.transaction(
            compare=[self.client.transactions.version(node_path) > 0],
            success=[self.client.transactions.delete(node_path)],
            failure=[],
        )
        if not status:
            raise JobLookupError(job_id)

    def remove_all_jobs(self):
        self.client.delete_prefix(self.path)

    def shutdown(self):
        self.client.close()

    def _reconstitute_job(self, job_state):
        job_state = job_state
        job = Job.__new__(Job)
        job.__setstate__(job_state)
        job._scheduler = self._scheduler
        job._jobstore_alias = self._alias
        return job

    def _get_jobs(self):
        jobs = []
        failed_job_ids = []
        all_ids = list(self.client.get_prefix(self.path))

        for doc, _ in all_ids:
            try:
                content = pickle.loads(doc)
                job_record = {
                    "next_run_time": content["next_run_time"],
                    "job": self._reconstitute_job(content["job_state"]),
                }
                jobs.append(job_record)
            except BaseException:
                content = pickle.loads(doc)
                failed_id = content["job_state"]["id"]
                failed_job_ids.append(failed_id)
                self._logger.exception(
                    'Unable to restore job "%s" -- removing it', failed_id
                )

        if failed_job_ids:
            for failed_id in failed_job_ids:
                self.remove_job(failed_id)
        paused_sort_key = datetime(9999, 12, 31, tzinfo=timezone.utc)
        return sorted(
            jobs,
            key=lambda job_record: job_record["job"].next_run_time or paused_sort_key,
        )

    def __repr__(self):
        self._logger.exception("<%s (client=%s)>", self.__class__.__name__, self.client)
        return f"<{self.__class__.__name__} (client={self.client})>"



"""Fields represent CronTrigger options which map to :class:`~datetime.datetime` fields."""

__all__ = (
    "MIN_VALUES",
    "MAX_VALUES",
    "DEFAULT_VALUES",
    "BaseField",
    "WeekField",
    "DayOfMonthField",
    "DayOfWeekField",
)

import re
from calendar import monthrange

from apscheduler.triggers.cron.expressions import (
    AllExpression,
    LastDayOfMonthExpression,
    MonthRangeExpression,
    RangeExpression,
    WeekdayPositionExpression,
    WeekdayRangeExpression,
)

MIN_VALUES = {
    "year": 1970,
    "month": 1,
    "day": 1,
    "week": 1,
    "day_of_week": 0,
    "hour": 0,
    "minute": 0,
    "second": 0,
}
MAX_VALUES = {
    "year": 9999,
    "month": 12,
    "day": 31,
    "week": 53,
    "day_of_week": 6,
    "hour": 23,
    "minute": 59,
    "second": 59,
}
DEFAULT_VALUES = {
    "year": "*",
    "month": 1,
    "day": 1,
    "week": "*",
    "day_of_week": "*",
    "hour": 0,
    "minute": 0,
    "second": 0,
}
SEPARATOR = re.compile(" *, *")


class BaseField:
    REAL = True
    COMPILERS = [AllExpression, RangeExpression]

    def __init__(self, name, exprs, is_default=False):
        self.name = name
        self.is_default = is_default
        self.compile_expressions(exprs)

    def get_min(self, dateval):
        return MIN_VALUES[self.name]

    def get_max(self, dateval):
        return MAX_VALUES[self.name]

    def get_value(self, dateval):
        return getattr(dateval, self.name)

    def get_next_value(self, dateval):
        smallest = None
        for expr in self.expressions:
            value = expr.get_next_value(dateval, self)
            if smallest is None or (value is not None and value < smallest):
                smallest = value

        return smallest

    def compile_expressions(self, exprs):
        self.expressions = []

        # Split a comma-separated expression list, if any
        for expr in SEPARATOR.split(str(exprs).strip()):
            self.compile_expression(expr)

    def compile_expression(self, expr):
        for compiler in self.COMPILERS:
            match = compiler.value_re.match(expr)
            if match:
                compiled_expr = compiler(**match.groupdict())

                try:
                    compiled_expr.validate_range(self.name)
                except ValueError as e:
                    raise ValueError(
                        f"Error validating expression {expr!r}: {e}"
                    ) from None

                self.expressions.append(compiled_expr)
                return

        raise ValueError(f'Unrecognized expression "{expr}" for field "{self.name}"')

    def __eq__(self, other):
        return (
            isinstance(self, self.__class__) and self.expressions == other.expressions
        )

    def __str__(self):
        expr_strings = (str(e) for e in self.expressions)
        return ",".join(expr_strings)

    def __repr__(self):
        return f"{self.__class__.__name__}('{self.name}', '{self}')"


class WeekField(BaseField):
    REAL = False

    def get_value(self, dateval):
        return dateval.isocalendar()[1]


class DayOfMonthField(BaseField):
    COMPILERS = BaseField.COMPILERS + [
        WeekdayPositionExpression,
        LastDayOfMonthExpression,
    ]

    def get_max(self, dateval):
        return monthrange(dateval.year, dateval.month)[1]


class DayOfWeekField(BaseField):
    REAL = False
    COMPILERS = BaseField.COMPILERS + [WeekdayRangeExpression]

    def get_value(self, dateval):
        return dateval.weekday()


class MonthField(BaseField):
    COMPILERS = BaseField.COMPILERS + [MonthRangeExpression]



import gc
import weakref
from datetime import datetime, timedelta
from functools import partial
from unittest.mock import MagicMock, patch

import pytest

from apscheduler.job import Job
from apscheduler.schedulers.base import BaseScheduler
from apscheduler.triggers.date import DateTrigger
from apscheduler.util import localize


def dummyfunc():
    pass


@pytest.fixture
def job(create_job):
    return create_job(func=dummyfunc)


@pytest.mark.parametrize("job_id", ["testid", None])
def test_constructor(job_id):
    with patch("apscheduler.job.Job._modify") as _modify:
        scheduler_mock = MagicMock(BaseScheduler)
        job = Job(scheduler_mock, id=job_id)
        assert job._scheduler is scheduler_mock
        assert job._jobstore_alias is None

        modify_kwargs = _modify.call_args[1]
        if job_id is None:
            assert len(modify_kwargs["id"]) == 32
        else:
            assert modify_kwargs["id"] == job_id


def test_modify(job):
    job.modify(bah=1, foo="x")
    job._scheduler.modify_job.assert_called_once_with(job.id, None, bah=1, foo="x")


def test_reschedule(job):
    job.reschedule("trigger", bah=1, foo="x")
    job._scheduler.reschedule_job.assert_called_once_with(
        job.id, None, "trigger", bah=1, foo="x"
    )


def test_pause(job):
    job.pause()
    job._scheduler.pause_job.assert_called_once_with(job.id, None)


def test_resume(job):
    job.resume()
    job._scheduler.resume_job.assert_called_once_with(job.id, None)


def test_remove(job):
    job.remove()
    job._scheduler.remove_job.assert_called_once_with(job.id, None)


def test_weakref(create_job):
    job = create_job(func=dummyfunc)
    ref = weakref.ref(job)
    del job
    gc.collect()
    assert ref() is None


def test_pending(job):
    """
    Tests that the "pending" property return True when _jobstore_alias is a string, ``False``
    otherwise.

    """
    assert job.pending

    job._jobstore_alias = "test"
    assert not job.pending


def test_get_run_times(create_job, timezone):
    run_time = localize(datetime(2010, 12, 13, 0, 8), timezone)
    expected_times = [run_time + timedelta(seconds=1), run_time + timedelta(seconds=2)]
    job = create_job(
        trigger="interval",
        trigger_args={"seconds": 1, "timezone": timezone, "start_date": run_time},
        next_run_time=expected_times[0],
        func=dummyfunc,
    )

    run_times = job._get_run_times(run_time)
    assert run_times == []

    run_times = job._get_run_times(expected_times[0])
    assert run_times == [expected_times[0]]

    run_times = job._get_run_times(expected_times[1])
    assert run_times == expected_times


def test_private_modify_bad_id(job):
    """Tests that only strings are accepted for job IDs."""
    del job.id
    exc = pytest.raises(TypeError, job._modify, id=3)
    assert str(exc.value) == "id must be a nonempty string"


def test_private_modify_id(job):
    """Tests that the job ID can't be changed."""
    exc = pytest.raises(ValueError, job._modify, id="alternate")
    assert str(exc.value) == "The job ID may not be changed"


def test_private_modify_bad_func(job):
    """Tests that given a func of something else than a callable or string raises a TypeError."""
    exc = pytest.raises(TypeError, job._modify, func=1)
    assert str(exc.value) == "func must be a callable or a textual reference to one"


def test_private_modify_func_ref(job):
    """Tests that the target callable can be given as a textual reference."""
    job._modify(func="tests.test_job:dummyfunc")
    assert job.func is dummyfunc
    assert job.func_ref == "tests.test_job:dummyfunc"


def test_private_modify_unreachable_func(job):
    """Tests that func_ref remains None if no reference to the target callable can be found."""
    func = partial(dummyfunc)
    job._modify(func=func)
    assert job.func is func
    assert job.func_ref is None


def test_private_modify_update_name(job):
    """Tests that the name attribute defaults to the function name."""
    del job.name
    job._modify(func=dummyfunc)
    assert job.name == "dummyfunc"


def test_private_modify_bad_args(job):
    """Tests that passing an argument list of the wrong type raises a TypeError."""
    exc = pytest.raises(TypeError, job._modify, args=1)
    assert str(exc.value) == "args must be a non-string iterable"


def test_private_modify_bad_kwargs(job):
    """Tests that passing an argument list of the wrong type raises a TypeError."""
    exc = pytest.raises(TypeError, job._modify, kwargs=1)
    assert str(exc.value) == "kwargs must be a dict-like object"


@pytest.mark.parametrize("value", [1, ""], ids=["integer", "empty string"])
def test_private_modify_bad_name(job, value):
    """
    Tests that passing an empty name or a name of something else than a string raises a TypeError.

    """
    exc = pytest.raises(TypeError, job._modify, name=value)
    assert str(exc.value) == "name must be a nonempty string"


@pytest.mark.parametrize("value", ["foo", 0, -1], ids=["string", "zero", "negative"])
def test_private_modify_bad_misfire_grace_time(job, value):
    """Tests that passing a misfire_grace_time of the wrong type raises a TypeError."""
    exc = pytest.raises(TypeError, job._modify, misfire_grace_time=value)
    assert (
        str(exc.value) == "misfire_grace_time must be either None or a positive integer"
    )


@pytest.mark.parametrize(
    "value", [None, "foo", 0, -1], ids=["None", "string", "zero", "negative"]
)
def test_private_modify_bad_max_instances(job, value):
    """Tests that passing a max_instances of the wrong type raises a TypeError."""
    exc = pytest.raises(TypeError, job._modify, max_instances=value)
    assert str(exc.value) == "max_instances must be a positive integer"


def test_private_modify_bad_trigger(job):
    """Tests that passing a trigger of the wrong type raises a TypeError."""
    exc = pytest.raises(TypeError, job._modify, trigger="foo")
    assert str(exc.value) == "Expected a trigger instance, got str instead"


def test_private_modify_bad_executor(job):
    """Tests that passing an executor of the wrong type raises a TypeError."""
    exc = pytest.raises(TypeError, job._modify, executor=1)
    assert str(exc.value) == "executor must be a string"


def test_private_modify_bad_next_run_time(job):
    """Tests that passing a next_run_time of the wrong type raises a TypeError."""
    exc = pytest.raises(TypeError, job._modify, next_run_time=1)
    assert str(exc.value) == "Unsupported type for next_run_time: int"


def test_private_modify_bad_argument(job):
    """Tests that passing an unmodifiable argument type raises an AttributeError."""
    exc = pytest.raises(AttributeError, job._modify, scheduler=1)
    assert (
        str(exc.value)
        == "The following are not modifiable attributes of Job: scheduler"
    )


def test_getstate(job):
    state = job.__getstate__()
    assert state == dict(
        version=1,
        trigger=job.trigger,
        executor="default",
        func="tests.test_job:dummyfunc",
        name=b"n\xc3\xa4m\xc3\xa9".decode("utf-8"),
        args=(),
        kwargs={},
        id=b"t\xc3\xa9st\xc3\xafd".decode("utf-8"),
        misfire_grace_time=1,
        coalesce=False,
        max_instances=1,
        next_run_time=None,
    )


def test_setstate(job, timezone):
    trigger = DateTrigger("2010-12-14 13:05:00", timezone)
    state = dict(
        version=1,
        scheduler=MagicMock(),
        jobstore=MagicMock(),
        trigger=trigger,
        executor="dummyexecutor",
        func="tests.test_job:dummyfunc",
        name="testjob.dummyfunc",
        args=[],
        kwargs={},
        id="other_id",
        misfire_grace_time=2,
        coalesce=True,
        max_instances=2,
        next_run_time=None,
    )
    job.__setstate__(state)
    assert job.id == "other_id"
    assert job.func == dummyfunc
    assert job.func_ref == "tests.test_job:dummyfunc"
    assert job.trigger == trigger
    assert job.executor == "dummyexecutor"
    assert job.args == []
    assert job.kwargs == {}
    assert job.name == "testjob.dummyfunc"
    assert job.misfire_grace_time == 2
    assert job.coalesce is True
    assert job.max_instances == 2
    assert job.next_run_time is None


def test_setstate_bad_version(job):
    """Tests that __setstate__ rejects state of higher version that it was designed to handle."""
    exc = pytest.raises(ValueError, job.__setstate__, {"version": 9999})
    assert "Job has version 9999, but only version" in str(exc.value)


def test_eq(create_job):
    job = create_job(func=lambda: None, id="foo")
    job2 = create_job(func=lambda: None, id="foo")
    job3 = create_job(func=lambda: None, id="bar")
    assert job == job2
    assert not job == job3
    assert not job == "foo"


def test_repr(job):
    assert repr(job) == "<Job (id=téstïd name=nämé)>"


@pytest.mark.parametrize(
    "status, expected_status",
    [
        ("scheduled", "next run at: 2011-04-03 18:40:00 CEST"),
        ("paused", "paused"),
        ("pending", "pending"),
    ],
    ids=["scheduled", "paused", "pending"],
)
@pytest.mark.parametrize("unicode", [False, True], ids=["nativestr", "unicode"])
def test_str(create_job, status, unicode, expected_status):
    job = create_job(func=dummyfunc)
    if status == "scheduled":
        job.next_run_time = job.trigger.run_date
    elif status == "pending":
        del job.next_run_time

    assert str(job) == (
        f"nämé (trigger: date[2011-04-03 18:40:00 CEST], {expected_status})"
    )



"""
Demonstrates how to use the background scheduler to schedule a job that executes on 3 second
intervals.
"""

import os
import time
from datetime import datetime

from apscheduler.schedulers.background import BackgroundScheduler


def tick():
    print(f"Tick! The time is: {datetime.now()}")


if __name__ == "__main__":
    scheduler = BackgroundScheduler()
    scheduler.add_job(tick, "interval", seconds=3)
    scheduler.start()
    print("Press Ctrl+{} to exit".format("Break" if os.name == "nt" else "C"))

    try:
        # This is here to simulate application activity (which keeps the main thread alive).
        while True:
            time.sleep(2)
    except (KeyboardInterrupt, SystemExit):
        # Not strictly necessary if daemonic mode is enabled but should be done if possible
        scheduler.shutdown()
